"""Methods to communicate with PyMC."""

# pylint: disable=protected-access
from copy import copy
from sys import modules

import numpy as np

try:
    from pytensor.tensor import matrix, TensorConstant
    from pytensor import function
    from pytensor.graph.basic import ancestors
    from pymc.pytensorf import compile_pymc, join_nonshared_inputs
    from pymc.util import is_transformed_name, get_untransformed_name
except ModuleNotFoundError:
    pass

from preliz.internal.distribution_helper import get_distributions


def back_fitting_pymc(prior, preliz_model, var_info):
    """
    Fit the samples from prior into user provided model's prior.
    from the perspective of ppe "prior" is actually an approximated posterior
    but from the users perspective is its prior.
    We need to "backfit" because we can not use arbitrary samples as priors.
    We need probability distributions.
    """
    new_priors = {}
    for rv_name, (_, size, *_) in var_info.items():
        if size > 1:
            params = []
            for i in range(size):
                opt_values = prior[rv_name][:, i]
                dist = preliz_model[rv_name]
                dist._fit_mle(opt_values)
                params.append(dist.params)
            dist._parametrization(*[np.array(x) for x in zip(*params)])
        else:
            opt_values = prior[rv_name]
            dist = preliz_model[rv_name]
            dist._fit_mle(opt_values)

        new_priors[rv_name] = dist

    return new_priors


def compile_mllk(model):
    """
    Compile the log-likelihood function for the model to be able to condition on both
    data and parameters.
    """
    obs_rvs = model.observed_RVs[0]
    old_y_value = model.rvs_to_values[obs_rvs]
    new_y_value = obs_rvs.type()
    model.rvs_to_values[obs_rvs] = new_y_value

    vars_ = model.value_vars
    initial_point = model.initial_point()

    [logp], raveled_inp = join_nonshared_inputs(
        point=initial_point, outputs=[model.datalogp], inputs=vars_
    )

    rv_logp_fn = compile_pymc([raveled_inp, new_y_value], logp)
    rv_logp_fn.trust_input = True

    def fmodel(params, obs):
        return -rv_logp_fn(params, obs).sum()

    return fmodel, old_y_value, obs_rvs


def get_initial_guess(model):
    """
    Get initial guess for optimization routine.
    """
    return np.concatenate([np.ravel(value) for value in model.initial_point().values()])


def extract_preliz_distributions(model):
    """
    Extract the corresponding PreliZ distributions from a PyMC model

    Parameters
    ----------
    model : a PyMC model

    Returns
    -------
    preliz_model : a dictionary of RVs names as keys and PreliZ distributions as values
    num_draws : the sample size of the observed
    """
    all_distributions = [
        dist
        for dist in modules["preliz.distributions"].__all__
        if dist not in ["Truncated", "Censored", "Hurdle", "Mixture"]
    ]
    pymc_to_preliz = dict(
        zip([dist.lower() for dist in all_distributions], get_distributions(all_distributions)),
    )

    preliz_model = {}
    for r_v in model.free_RVs:
        dist_name = (
            r_v.owner.op.name if r_v.owner.op.name else str(r_v.owner.op).split("RV", 1)[0].lower()
        )
        dist = copy(pymc_to_preliz[dist_name])
        preliz_model[r_v.name] = dist

    return preliz_model


def retrieve_variable_info(model):
    """
    Get the shape, size, transformation and parents of each free random variable in a PyMC model.
    """

    var_info = {}
    initial_point = model.initial_point()
    for v_var in model.value_vars:
        name = v_var.name
        rvs = model.values_to_rvs[v_var]
        nc_parents = non_constant_parents(rvs, model)
        idx_parents = []
        if nc_parents:
            idx_parents = [model.free_RVs.index(var_) for var_ in nc_parents]

        if is_transformed_name(name):
            name = get_untransformed_name(name)
            x_var = matrix(f"{name}_transformed")
            z_var = model.rvs_to_transforms[rvs].backward(x_var)
            transformation = function(inputs=[x_var], outputs=z_var)
        else:
            transformation = None

        var_info[name] = (
            initial_point[v_var.name].shape,
            initial_point[v_var.name].size,
            transformation,
            idx_parents,
        )

    num_draws = model.observed_RVs[0].eval().size

    return var_info, num_draws


def unravel_projection(prior_array, var_info, iterations):
    size = 0
    prior_dict = {}
    for key, values in var_info.items():
        shape, new_size, transformation, _ = values
        vector = prior_array[:, size : size + new_size]
        if transformation is not None:
            vector = transformation(vector)
        prior_dict[key] = vector.reshape(iterations, *shape).squeeze()
        size += new_size

    return prior_dict


def write_pymc_string(new_priors, var_info):
    """
    Return a string with the new priors for the PyMC model.
    So the user can copy and paste, ideally with none to minimal changes.
    """

    header = "with pm.Model() as model:\n"
    variables = []
    names = list(new_priors.keys())
    for key, value in new_priors.items():
        idxs = var_info[key][-1]
        if idxs:
            for i in idxs:
                nkey = names[i]
                cp_dist = copy(new_priors[nkey])
                cp_dist._fit_moments(np.mean(value.mean()), np.mean(value.std()))

                dist_name, dist_params = repr(cp_dist).split("(")
                size = var_info[nkey][1]
                if size > 1:
                    dist_params = dist_params.split(")")[0]
                    variables[
                        i
                    ] = f'    {nkey:} = pm.{dist_name}("{nkey}", {dist_params}, shape={size})\n'
                else:
                    variables[i] = f'    {nkey:} = pm.{dist_name}("{nkey}", {dist_params}\n'

        else:
            dist_name, dist_params = repr(value).split("(")
            size = var_info[key][1]
            if size > 1:
                dist_params = dist_params.split(")")[0]
                variables.append(
                    f'    {key:} = pm.{dist_name}("{key}", {dist_params}, shape={size})\n'
                )
            else:
                variables.append(f'    {key:} = pm.{dist_name}("{key}", {dist_params}\n')

    return "".join([header] + variables)


def non_constant_parents(rvs, model):
    """Find the parents of a variable that are not constant."""
    parents = []
    for variable in rvs.get_parents()[0].inputs[2:]:
        if not isinstance(variable, TensorConstant):
            for free_rv in model.free_RVs:
                if free_rv in list(ancestors([variable])) and free_rv not in parents:
                    parents.append(free_rv)
    return parents



from preliz.internal.distribution_helper import process_extra


def test_process_extra():
    ref0 = {"TruncatedNormal": {"lower": -3, "upper": 3}}
    ref1 = {"StudentT": {"nu": 3.4}, "Normal": {"mu": 3.0}}

    assert process_extra("TruncatedNormal(lower=-3, upper=3)") == ref0
    assert process_extra("StudentT(nu=3.4),Normal(mu=3)") == ref1



"""Tests plot_utils functions."""
# pylint: disable=redefined-outer-name
import pytest
import matplotlib.pyplot as plt
import numpy as np

import preliz as pz
from test_helper import run_notebook


@pytest.fixture(scope="function")
def two_dist():
    return pz.Beta(2, 6), pz.Poisson(4.5)


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"moments": "mdsk"},
        {"moments": "mdsk", "legend": "title"},
        {"pointinterval": True},
        {"pointinterval": True, "levels": [0.1, 0.9]},
        {"pointinterval": True, "interval": "eti", "levels": [0.9]},
        {"pointinterval": True, "interval": "quantiles"},
        {"pointinterval": True, "interval": "quantiles", "levels": [0.1, 0.5, 0.9]},
        {"support": "restricted"},
        {"color": "C1", "alpha": 0.1},
        {"figsize": (4, 4)},
        {"ax": plt.subplots()[1]},
    ],
)
def test_continuous_plot_pdf_cdf_ppf(two_dist, kwargs):
    for a_dist in two_dist:
        a_dist.plot_pdf(**kwargs)
        a_dist.plot_cdf(**kwargs)
        kwargs.pop("support", None)
        a_dist.plot_ppf(**kwargs)


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"marginals": True},
        {"pointinterval": True},
        {"pointinterval": True, "levels": [0.1, 0.9]},
        {"pointinterval": True, "interval": "eti", "levels": [0.9]},
        {"pointinterval": True, "interval": "quantiles"},
        {"pointinterval": True, "interval": "quantiles", "levels": [0.1, 0.5, 0.9]},
        {"support": "restricted"},
        {"figsize": (4, 4)},
    ],
)
def test_dirichlet_plot(kwargs):
    a_dist = pz.Dirichlet([2, 1, 2])
    a_dist.plot_pdf(**kwargs)
    kwargs.pop("marginals", None)
    a_dist.plot_cdf(**kwargs)
    kwargs.pop("support", None)
    a_dist.plot_ppf(**kwargs)


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"marginals": True},
        {"pointinterval": True},
        {"pointinterval": True, "levels": [0.1, 0.9]},
        {"pointinterval": True, "interval": "eti", "levels": [0.9]},
        {"pointinterval": True, "interval": "quantiles"},
        {"pointinterval": True, "interval": "quantiles", "levels": [0.1, 0.5, 0.9]},
        {"support": "restricted"},
        {"figsize": (4, 4)},
    ],
)
def test_mvnormal_plot(kwargs):
    a_dist = pz.MvNormal(np.zeros(2), np.eye(2))
    a_dist.plot_pdf(**kwargs)
    kwargs.pop("marginals", None)
    a_dist.plot_cdf(**kwargs)
    kwargs.pop("support", None)
    a_dist.plot_ppf(**kwargs)


def test_plot_interactive(capsys):
    pz.Poisson(4.5).plot_interactive()
    captured = capsys.readouterr()
    assert "RuntimeError" in captured.out
    run_notebook("plot_interactive.ipynb")


@pytest.fixture
def sample_ax():
    return plt.subplot()


def test_plot_references(sample_ax):
    # Test with a dictionary of references
    references_dict = {"Ref1": 0.5, "Ref2": 1.0, "Ref3": 1.5}
    pz.internal.plot_helper.plot_references(references_dict, sample_ax)

    lines = sample_ax.lines
    texts = sample_ax.texts

    assert len(lines) == len(texts) == len(references_dict)

    # Test with a list of references
    sample_ax.clear()
    references_list = [0.5, 1.0, 1.5]
    pz.internal.plot_helper.plot_references(references_list, sample_ax)

    lines = sample_ax.lines

    assert len(lines) == len(references_list)

    # Test with a single reference value
    sample_ax.clear()
    reference_single = 0.5
    pz.internal.plot_helper.plot_references(reference_single, sample_ax)

    lines = sample_ax.lines

    assert len(lines) == 1

    # Test with None input
    sample_ax.clear()
    references_none = None
    pz.internal.plot_helper.plot_references(references_none, sample_ax)

    lines = sample_ax.lines
    texts = sample_ax.texts

    assert len(lines) == len(texts) == 0



# pylint: disable=arguments-differ
import numpy as np

from preliz.distributions.distributions import DistributionTransformer
from preliz.internal.distribution_helper import all_not_none, num_skewness, num_kurtosis


class Truncated(DistributionTransformer):
    r"""
    Truncated distribution

    This is not a distribution per se, but a modifier of univariate distributions.

    Given a base distribution with cumulative distribution function (CDF) and
    probability density mass/function (PDF). The pdf of a Truncated distribution is:

    .. math::

        \begin{cases}
            0 & \text{for } x < \text{lower}, \\
            \frac{\text{PDF}(x)}{\text{CDF}(upper) - \text{CDF}(lower)}
            & \text{for } \text{lower} <= x <= \text{upper}, \\
            0 & \text{for } x > \text{upper},
        \end{cases}

    .. plot::
        :context: close-figs

        
        from preliz import Gamma, Truncated, style
        style.use('preliz-doc')
        Truncated(Gamma(mu=2, sigma=1), 1, 4.5).plot_pdf()
        Gamma(mu=2, sigma=1).plot_pdf()
        

    Parameters
    ----------
    dist: PreliZ distribution
        Univariate PreliZ distribution which will be truncated.
    lower: float or int
        Lower (left) truncation point. Use np.inf for no truncation.
    upper: float or int
        Upper (right) truncation point. Use np.inf for no truncation.
    """

    def __init__(self, dist, lower=None, upper=None, **kwargs):
        self.dist = dist
        super().__init__()
        self._parametrization(lower, upper, **kwargs)

    def _parametrization(self, lower=None, upper=None, **kwargs):
        dist_params = []
        if not kwargs:
            if hasattr(self.dist, "params"):
                kwargs = dict(zip(self.dist.param_names, self.dist.params))
            else:
                kwargs = dict(zip(self.dist.param_names, [None] * len(self.dist.param_names)))

        for key, value in kwargs.items():
            dist_params.append(value)
            setattr(self, key, value)

        if upper is None:
            self.upper = np.inf
        else:
            self.upper = upper

        if lower is None:
            self.lower = -np.inf
        else:
            self.lower = lower

        self.params = (*dist_params, self.lower, self.upper)
        self.param_names = (*self.dist.param_names, "lower", "upper")
        if all_not_none(*dist_params):
            self.dist._parametrization(**kwargs)
            self.is_frozen = True

        self.support = (
            np.maximum(self.dist.support[0], self.lower),
            np.minimum(self.dist.support[1], self.upper),
        )
        self.params_support = (*self.dist.params_support, self.dist.support, self.dist.support)

    def mean(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        if self.kind == "discrete":
            return np.sum(x_values * pdf)
        else:
            return np.trapz(x_values * pdf, x_values)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        if self.kind == "discrete":
            return np.sum((x_values - self.mean()) ** 2 * pdf)
        else:
            return np.trapz((x_values - self.mean()) ** 2 * pdf, x_values)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return num_skewness(self)

    def kurtosis(self):
        return num_kurtosis(self)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return self.ppf(random_state.uniform(size=size))

    def pdf(self, x):
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        lower = adjust_lower(self.kind, self.lower)
        lcdf = self.dist.cdf(lower)
        vals = (self.dist.cdf(x) - lcdf) / (self.dist.cdf(self.upper) - lcdf)
        return np.where(x < lower, 0, np.where(x > self.upper, 1, vals))

    def ppf(self, q):
        q = np.asarray(q)
        lower = adjust_lower(self.kind, self.lower)
        lcdf = self.dist.cdf(lower)
        vals = self.dist.ppf(lcdf + q * (self.dist.cdf(self.upper) - lcdf))
        return np.where((q < 0) | (q > 1), np.nan, vals)

    def logpdf(self, x):
        x = np.asarray(x)
        lower = adjust_lower(self.kind, self.lower)
        vals = self.dist.logpdf(x) - np.log(self.dist.cdf(self.upper) - self.dist.cdf(lower))
        return np.where((x < self.lower) | (x > self.upper), -np.inf, vals)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        if self.kind == "discrete":
            return -np.sum(np.exp(logpdf) * logpdf)
        else:
            return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def _neg_logpdf(self, x):
        return -self.logpdf(x).sum()

    def _fit_moments(self, mean, sigma):
        self.dist._fit_moments(mean, sigma)
        self._parametrization(**dict(zip(self.dist.param_names, self.dist.params)))


def adjust_lower(kind, lower):
    if kind == "discrete":
        lower -= 1
    return lower



import numpy as np


from preliz.internal.plot_helper import repr_to_matplotlib
from preliz.internal.distribution_helper import get_distributions
from preliz.unidimensional import mle


def back_fitting_ppa(model, subset, new_families=True):
    """
    Use MLE to fit a subset of the prior samples to the marginal prior distributions
    """
    string = "Your selection is consistent with the priors (original families):\n"

    for name, dist in model.items():
        dist._fit_mle(subset[name])
        string += f"{name} = {repr_to_matplotlib(dist)}\n"

    if new_families:
        string += "\nYour selection is consistent with the priors (new families):\n"

        # We should store this in a central place
        # So we use the same families for other functions
        common_cont = ["Gamma", "Exponential"]
        common_disc = ["Poisson", "NegativeBinomial"]
        for name, dist in model.items():
            if dist.kind == "continuous":
                distributions = get_distributions(set([dist.__class__.__name__] + common_cont))
            elif dist.kind == "discrete":
                distributions = get_distributions(set([dist.__class__.__name__] + common_disc))
            idx, _ = mle(distributions, subset[name], plot=False)
            string += f"{name} = {repr_to_matplotlib(distributions[idx[0]])}\n"

    return string, np.concatenate([dist.params for dist in model.values()])


def select_prior_samples(selected, prior_samples, model):
    """
    Given a selected set of prior predictive samples pick the corresponding
    prior samples.
    """
    subsample = {rv: prior_samples[rv][selected] for rv in model.keys()}

    return subsample



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.optimization import optimize_ml
from ..internal.special import (
    betaln,
    betainc,
    betaincinv,
    digamma,
    cdf_bounds,
    ppf_bounds_cont,
    xlogy,
)


class BetaScaled(Continuous):
    r"""
    Scaled Beta distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \alpha, \beta) =
           \frac{(x-\text{lower})^{\alpha - 1} (\text{upper} - x)^{\beta - 1}}
           {(\text{upper}-\text{lower})^{\alpha+\beta-1} B(\alpha, \beta)}

    .. plot::
        :context: close-figs


        from preliz import BetaScaled, style
        style.use('preliz-doc')
        alphas = [2, 2]
        betas = [2, 5]
        lowers = [-0.5, -1]
        uppers = [1.5, 2]
        for alpha, beta, lower, upper in zip(alphas, betas, lowers, uppers):
            BetaScaled(alpha, beta, lower, upper).plot_pdf()

    ========  ==============================================================
    Support   :math:`x \in (lower, upper)`
    Mean      :math:`\dfrac{\alpha}{\alpha + \beta} (upper-lower) + lower`
    Variance  :math:`\dfrac{\alpha \beta}{(\alpha+\beta)^2(\alpha+\beta+1)} (upper-lower)`
    ========  ==============================================================

    Parameters
    ----------
    alpha : float
        alpha  > 0
    beta : float
        beta  > 0
    lower: float
        Lower limit.
    upper: float
        Upper limit (upper > lower).
    """

    def __init__(self, alpha=None, beta=None, lower=0, upper=1):
        super().__init__()
        self.alpha = alpha
        self.beta = beta
        self.lower = lower
        self.upper = upper
        self.support = (lower, upper)
        self._parametrization(self.alpha, self.beta, self.lower, self.upper)

    def _parametrization(self, alpha=None, beta=None, lower=None, upper=None):
        self.param_names = ("alpha", "beta", "lower", "upper")
        self.params_support = ((eps, np.inf), (eps, np.inf), (-np.inf, np.inf), (-np.inf, np.inf))
        if all_not_none(alpha, beta):
            self._update(alpha, beta, lower, upper)

    def _update(self, alpha, beta, lower=None, upper=None):
        if lower is not None:
            self.lower = np.float64(lower)
        if upper is not None:
            self.upper = np.float64(upper)

        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.params = (self.alpha, self.beta, self.lower, self.upper)
        self.support = self.lower, self.upper
        self.is_frozen = True

    def _from_mu_sigma(self, mu, sigma):
        nu = mu * (1 - mu) / sigma**2 - 1
        alpha = mu * nu
        beta = (1 - mu) * nu
        return alpha, beta

    def _from_mu_nu(self, mu, nu):
        alpha = mu * nu
        beta = (1 - mu) * nu
        return alpha, beta

    def _to_mu_sigma(self, alpha, beta):
        alpha_plus_beta = alpha + beta
        mu = alpha / alpha_plus_beta
        sigma = (alpha * beta) ** 0.5 / alpha_plus_beta / (alpha_plus_beta + 1) ** 0.5
        return mu, sigma

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta, self.lower, self.upper)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta, self.lower, self.upper)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta, self.lower, self.upper)

    def entropy(self):
        return nb_entropy(self.alpha, self.beta, self.lower, self.upper)

    def mean(self):
        return (self.alpha * self.upper + self.beta * self.lower) / (self.alpha + self.beta)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return (
            (self.alpha * self.beta)
            / ((self.alpha + self.beta) ** 2 * (self.alpha + self.beta + 1))
            * (self.lower - self.upper) ** 2
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        if self.alpha == self.beta:
            return np.zeros_like(self.alpha)
        else:
            psc = self.alpha + self.beta
            return (2 * (self.beta - self.alpha) * np.sqrt(psc + 1)) / (
                (psc + 2) * np.sqrt(self.alpha * self.beta)
            )

    def kurtosis(self):
        psc = self.alpha + self.beta
        prod = self.alpha * self.beta
        return (
            6
            * (np.abs(self.alpha - self.beta) ** 2 * (psc + 1) - prod * (psc + 2))
            / (prod * (psc + 2) * (psc + 3))
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return (
            random_state.beta(self.alpha, self.beta, size) * (self.upper - self.lower) + self.lower
        )

    def _fit_moments(self, mean, sigma):
        mean = (mean - self.lower) / (self.upper - self.lower)
        sigma = sigma / (self.upper - self.lower)
        kappa = mean * (1 - mean) / sigma**2 - 1
        alpha = max(0.5, kappa * mean)
        beta = max(0.5, kappa * (1 - mean))
        self._update(alpha, beta)

    def _fit_mle(self, sample):
        self._update(None, None, np.min(sample), np.max(sample))
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, alpha, beta, lower, upper):
    prob = betainc(alpha, beta, (x - lower) / (upper - lower))
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, alpha, beta, lower, upper):
    x_val = betaincinv(alpha, beta, q) * (upper - lower) + lower
    return ppf_bounds_cont(x_val, q, lower, upper)


# @nb.njit(cache=True)
def nb_entropy(alpha, beta, lower, upper):
    psc = alpha + beta
    return (
        betaln(alpha, beta)
        - (alpha - 1) * digamma(alpha)
        - (beta - 1) * digamma(beta)
        + (psc - 2) * digamma(psc)
        + np.log(upper - lower)
    )


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta, lower, upper):
    if x < lower or x > upper:
        return -np.inf
    else:
        return (xlogy((alpha - 1), (x - lower)) + xlogy((beta - 1), (upper - x))) - (
            xlogy((alpha + beta - 1), (upper - lower)) + betaln(alpha, beta)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta, lower, upper):
    return -(nb_logpdf(x, alpha, beta, lower, upper)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np
from scipy.special import bdtr, bdtrik  # pylint: disable=no-name-in-module

from .distributions import Discrete
from ..internal.optimization import optimize_moments, optimize_ml
from ..internal.distribution_helper import eps, all_not_none
from ..internal.special import cdf_bounds, ppf_bounds_disc, gammaln


class ZeroInflatedBinomial(Discrete):
    R"""
    Zero-inflated Binomial distribution.

    The pmf of this distribution is

    .. math::

        f(x \mid \psi, n, p) = \left\{ \begin{array}{l}
            (1-\psi) + \psi (1-p)^{n}, \text{if } x = 0 \\
            \psi {n \choose x} p^x (1-p)^{n-x}, \text{if } x=1,2,3,\ldots,n
            \end{array} \right.

    .. plot::
        :context: close-figs

        
        from preliz import ZeroInflatedBinomial, style
        style.use('preliz-doc')
        ns = [10, 20]
        ps = [0.5, 0.7]
        psis = [0.7, 0.4]
        for n, p, psi in zip(ns, ps, psis):
            ZeroInflatedBinomial(psi, n, p).plot_pdf(support=(0,25))

    ========  ==========================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\psi n p`
    Variance  :math:`\psi n p (1 - p) + n^2 p^2 (\psi - \psi^2)`
    ========  ==========================

    Parameters
    ----------
    psi : float
        Expected proportion of Binomial variates (0 < psi < 1)
    n : int
        Number of Bernoulli trials (n >= 0).
    p : float
        Probability of success in each trial (0 < p < 1).
    """

    def __init__(self, psi=None, n=None, p=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(psi, n, p)

    def _parametrization(self, psi=None, n=None, p=None):
        self.psi = psi
        self.n = n
        self.p = p
        self.params = (self.psi, self.n, self.p)
        self.param_names = ("psi", "n", "p")
        self.params_support = ((eps, 1 - eps), (eps, np.inf), (eps, 1 - eps))
        if all_not_none(psi, n, p):
            self._update(psi, n, p)

    def _update(self, psi, n, p):
        self.psi = np.float64(psi)
        self.n = np.int64(n)
        self.p = np.float64(p)
        self.params = (self.psi, self.n, self.p)
        if self.psi == 0:
            self.support = (0, 0)
        else:
            self.support = (0, self.n)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(self.psi, self.n, x, self.p))

    def cdf(self, x):
        return nb_cdf(x, self.psi, self.n, self.p, self.support[0], self.support[1])

    def ppf(self, q):
        return nb_ppf(q, self.psi, self.n, self.p, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(self.psi, self.n, x, self.p)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(self.psi, self.n, x, self.p)

    def entropy(self):
        binomial_entropy = 0.5 * np.log(2 * np.pi * np.e * self.n * self.p * (1 - self.p))
        if self.psi == 1:
            return binomial_entropy
        else:
            zero_entropy = -(1 - self.psi) * np.log(1 - self.psi) - self.psi * np.log(self.psi)
            return (1 - self.psi) * zero_entropy + self.psi * binomial_entropy

    def mean(self):
        return self.psi * self.n * self.p

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.mean() * (1 - self.p) + self.n**2 * self.p**2 * (self.psi - self.psi**2)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        # implement skewness
        return np.nan

    def kurtosis(self):
        # implement kurtosis
        return np.nan

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        zeros = random_state.uniform(size=size) > (1 - self.psi)
        binomial = random_state.binomial(self.n, self.p, size=size)
        return zeros * binomial

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


# @nb.jit
# bdtr not supported by numba
def nb_cdf(x, psi, n, p, lower, upper):
    x = np.asarray(x)
    b_prob = np.asarray(bdtr(x, n, p))
    prob = (1 - psi) + psi * b_prob
    return cdf_bounds(prob, x, lower, upper)


# @nb.jit
def nb_ppf(q, psi, n, p, lower, upper):
    q = np.asarray(q)
    n_vals = np.ceil(bdtrik(q, n, p))
    x_vals = (1 - psi) + psi * n_vals
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(psi, n, y, p):
    if y < 0:
        return -np.inf
    elif y > n:
        return -np.inf
    elif y == 0:
        return np.log((1 - psi) + psi * (1 - p) ** n)
    else:
        return (
            np.log(psi)
            + gammaln(n + 1)
            - (gammaln(y + 1) + gammaln(n - y + 1))
            + y * np.log(p)
            + (n - y) * np.log1p(-p)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(psi, n, y, p):
    return -(nb_logpdf(psi, n, y, p)).sum()



"""
PreliZ.

Exploring and eliciting probability distributions
"""
from os import path as os_path

from matplotlib import rcParams as mpl_rcParams
from matplotlib import style

from .distributions import *
from .predictive import *
from .unidimensional import *
from .multidimensional import *
from .internal.rcparams import rc_context, rcParams


__version__ = "0.15.0"


# Allow legend outside plot in maxent to be included in the saved figure
mpl_rcParams["savefig.bbox"] = "tight"


# add PreliZ's styles to matplotlib's styles
_preliz_style_path = os_path.join(os_path.dirname(__file__), "styles")
style.core.USER_LIBRARY_PATHS.append(_preliz_style_path)
style.core.reload_library()

# clean namespace
del os_path, mpl_rcParams, _preliz_style_path



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, to_precision, from_precision, all_not_none
from ..internal.special import half_erf, erfinv, ppf_bounds_cont


class HalfNormal(Continuous):
    r"""
    HalfNormal Distribution

    The pdf of this distribution is

    .. math::

       f(x \mid \sigma) =
           \sqrt{\frac{2}{\pi\sigma^2}}
           \exp\left(\frac{-x^2}{2\sigma^2}\right)

    .. plot::
        :context: close-figs


        from preliz import HalfNormal, style
        style.use('preliz-doc')
        for sigma in [0.4,  2.]:
            HalfNormal(sigma).plot_pdf(support=(0,5))

    ========  ==========================================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`\dfrac{\sigma \sqrt{2}}{\sqrt{\pi}}`
    Variance  :math:`\sigma^2\left(1 - \dfrac{2}{\pi}\right)`
    ========  ==========================================

    HalfNormal distribution has 2 alternative parameterizations. In terms of sigma (standard
    deviation) or tau (precision).

    The link between the 2 alternatives is given by

    .. math::

        \tau = \frac{1}{\sigma^2}

    Parameters
    ----------
    sigma : float
        Scale parameter :math:`\sigma` (``sigma`` > 0).
    tau : float
        Precision :math:`\tau` (``tau`` > 0).
    """

    def __init__(self, sigma=None, tau=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(sigma, tau)

    def _parametrization(self, sigma=None, tau=None):
        if all_not_none(sigma, tau):
            raise ValueError("Incompatible parametrization. Either use sigma or tau.")

        self.param_names = ("sigma",)
        self.params_support = ((eps, np.inf),)

        if tau is not None:
            sigma = from_precision(tau)
            self.param_names = ("tau",)

        self.sigma = sigma
        self.tau = tau
        if self.sigma is not None:
            self._update(self.sigma)

    def _update(self, sigma):
        self.sigma = np.float64(sigma)
        self.tau = to_precision(self.sigma)

        if self.param_names[0] == "sigma":
            self.params = (self.sigma,)
        elif self.param_names[0] == "tau":
            self.params = (self.tau,)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.sigma))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.sigma, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.sigma)

    def entropy(self):
        return nb_entropy(self.sigma)

    def mean(self):
        return self.sigma * 0.7978845608028655

    def mode(self):
        return np.zeros_like(self.sigma)

    def median(self):
        return self.sigma * 0.6744897501960818

    def var(self):
        return self.sigma**2 * 0.3633802276324186

    def std(self):
        return self.sigma * 0.6028102749890869

    def skewness(self):
        return 0.9952717464311565

    def kurtosis(self):
        return 0.8691773036059736

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return np.abs(random_state.normal(0, self.sigma, size))

    def _fit_moments(self, mean, sigma):  # pylint: disable=unused-argument
        self._update(sigma / (1 - 2 / np.pi) ** 0.5)

    def _fit_mle(self, sample):
        self._update(nb_fit_mle(sample))


@nb.njit(cache=True)
def nb_cdf(x, sigma):
    return half_erf(x / (sigma * 2**0.5))


@nb.njit(cache=True)
def nb_ppf(q, sigma, lower, upper):
    x_vals = np.asarray(sigma * 2**0.5 * erfinv(q))
    return ppf_bounds_cont(x_vals, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(sigma):
    return 0.5 * np.log(np.pi * sigma**2.0 / 2.0) + 0.5


@nb.njit(cache=True)
def nb_fit_mle(sample):
    return np.mean(sample**2) ** 0.5


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, sigma):
    if x < 0:
        return -np.inf
    else:
        return np.log(np.sqrt(2 / np.pi)) + np.log(1 / sigma) - 0.5 * ((x / sigma) ** 2)


@nb.njit(cache=True)
def nb_neg_logpdf(x, sigma):
    return -(nb_logpdf(x, sigma)).sum()



# pylint: disable=too-many-lines
# pylint: disable=too-many-instance-attributes
# pylint: disable=invalid-name
# pylint: disable=attribute-defined-outside-init
"""
Continuous multivariate probability distributions.
"""
from copy import copy

import numpy as np

try:
    from ipywidgets import interactive, widgets
except ImportError:
    pass
from scipy import stats

from .beta import Beta  # pylint: disable=no-name-in-module
from .normal import Normal  # pylint: disable=no-name-in-module
from .distributions_multivariate import Continuous
from ..internal.distribution_helper import all_not_none
from ..internal.plot_helper_multivariate import plot_dirichlet, plot_mvnormal
from ..internal.plot_helper import check_inside_notebook, get_slider

eps = np.finfo(float).eps


class Dirichlet(Continuous):
    r"""
    Dirichlet distribution.

    .. math::

       f(\mathbf{x}|\mathbf{a}) =
           \frac{\Gamma(\sum_{i=1}^k a_i)}{\prod_{i=1}^k \Gamma(a_i)}
           \prod_{i=1}^k x_i^{a_i - 1}

    .. plot::
        :context: close-figs


        import matplotlib.pyplot as plt
        from preliz import Dirichlet
        _, axes = plt.subplots(2, 2)
        alphas = [[0.5, 0.5, 0.5], [1, 1, 1], [5, 5, 5], [5, 2, 1]]
        for alpha, ax in zip(alphas, axes.ravel()):
            pz.Dirichlet(alpha).plot_pdf(marginals=False, ax=ax)


    ========  ===============================================
    Support   :math:`x_i \in (0, 1)` for :math:`i \in \{1, \ldots, K\}`
              such that :math:`\sum x_i = 1`
    Mean      :math:`\dfrac{a_i}{\sum a_i}`
    Variance  :math:`\dfrac{a_i - \sum a_0}{a_0^2 (a_0 + 1)}`
              where :math:`a_0 = \sum a_i`
    ========  ===============================================

    Parameters
    ----------
    alpha : array of floats
        Concentration parameter (alpha > 0).
    """

    def __init__(self, alpha=None):
        super().__init__()
        self.dist = copy(stats.dirichlet)
        self.marginal = Beta
        self.support = (eps, 1 - eps)
        self._parametrization(alpha)

    def _parametrization(self, alpha=None):
        self.param_names = ("alpha",)
        self.params_support = ((eps, np.inf),)

        self.alpha = alpha
        if alpha is not None:
            self._update(alpha)

    def _get_frozen(self):
        frozen = None
        if all_not_none(self):
            frozen = self.dist(self.alpha)
        return frozen

    def _update(self, alpha):
        self.alpha = np.array(alpha, dtype=float)
        self.params = (self.alpha,)
        self._update_rv_frozen()

    def _fit_mle(self, sample, **kwargs):
        raise NotImplementedError

    def plot_pdf(
        self,
        marginals=True,
        pointinterval=False,
        interval=None,
        levels=None,
        support="full",
        legend="title",
        figsize=None,
        ax=None,
    ):
        """
        Plot the pdf of the marginals or the joint pdf of the simplex.
        The joint representation is only available for a dirichlet with an alpha of length 3.

        Parameters
        ----------
        marginals : True
            Defaults to True, plot the marginal distributions, if False plot the joint distribution
            (only valid for an alpha of length 3).
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantile ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axis
            Axis to plot on

        Returns
        -------
        ax : matplotlib axis
        """
        return plot_dirichlet(
            self, "pdf", marginals, pointinterval, interval, levels, support, legend, figsize, ax
        )

    def plot_cdf(
        self,
        pointinterval=False,
        interval=None,
        levels=None,
        support="full",
        legend="title",
        figsize=None,
        ax=None,
    ):
        """
        Plot the cumulative distribution function.

        Parameters
        ----------
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantile ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axis
            Axis to plot on

        Returns
        -------
        ax : matplotlib axis
        """
        return plot_dirichlet(
            self, "cdf", "marginals", pointinterval, interval, levels, support, legend, figsize, ax
        )

    def plot_ppf(
        self,
        pointinterval=False,
        interval=None,
        levels=None,
        legend="title",
        figsize=None,
        ax=None,
    ):
        """
        Plot the quantile function.

        Parameters
        ----------
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantile ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axis
            Axis to plot on

        Returns
        -------
        ax : matplotlib axis
        """
        return plot_dirichlet(
            self, "ppf", "marginals", pointinterval, interval, levels, None, legend, figsize, ax
        )

    def plot_interactive(
        self,
        kind="pdf",
        xy_lim="both",
        pointinterval=True,
        interval=None,
        levels=None,
        legend="title",
        figsize=None,
    ):
        """
        Interactive exploration of parameters

        Parameters
        ----------
        kind : str:
            Type of plot. Available options are `pdf`, `cdf` and `ppf`.
        xy_lim : str or tuple
            Set the limits of the x-axis and/or y-axis.
            Defaults to `"both"`, the limits of both axes are fixed for all subplots.
            Use `"auto"` for automatic rescaling of x-axis and y-axis.
            Or set them manually by passing a tuple of 4 elements,
            the first two for x-axis, the last two for y-axis. The tuple can have `None`.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False.
            If `True` the default is to plot the median and two inter-quantiles ranges.
        interval : str
            Type of interval. Available options are the highest density interval `"hdi"`,
            equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        """

        check_inside_notebook()

        args = dict(zip(self.param_names, self.params))
        self.__init__(**args)  # pylint: disable=unnecessary-dunder-call
        if kind == "pdf":
            w_checkbox_marginals = widgets.Checkbox(
                value=True,
                description="marginals",
                disabled=False,
                indent=False,
            )
            plot_widgets = {"marginals": w_checkbox_marginals}
        else:
            plot_widgets = {}
        for index, dim in enumerate(self.params[0]):
            plot_widgets[f"alpha-{index + 1}"] = get_slider(
                f"alpha-{index + 1}", dim, *self.params_support[0]
            )

        def plot(**args):
            if kind == "pdf":
                marginals = args.pop("marginals")
            params = {"alpha": np.asarray(list(args.values()), dtype=float)}
            self.__init__(**params)  # pylint: disable=unnecessary-dunder-call
            if kind == "pdf":
                plot_dirichlet(
                    self,
                    "pdf",
                    marginals,
                    pointinterval,
                    interval,
                    levels,
                    "full",
                    legend,
                    figsize,
                    None,
                    xy_lim,
                )
            elif kind == "cdf":
                plot_dirichlet(
                    self,
                    "cdf",
                    "marginals",
                    pointinterval,
                    interval,
                    levels,
                    "full",
                    figsize,
                    None,
                    xy_lim,
                )
            elif kind == "ppf":
                plot_dirichlet(
                    self,
                    "cdf",
                    "marginals",
                    pointinterval,
                    interval,
                    levels,
                    None,
                    figsize,
                    None,
                    xy_lim,
                )

        return interactive(plot, **plot_widgets)


class MvNormal(Continuous):
    r"""
    Multivariate Normal distribution.

    .. math::

       f(x \mid \mu, T) =
           \frac{|T|^{1/2}}{(2\pi)^{k/2}}
           \exp\left\{ -\frac{1}{2} (x-\mu)^{\prime} T (x-\mu) \right\}

    .. plot::
        :context: close-figs


        import matplotlib.pyplot as plt
        from preliz import MvNormal
        _, axes = plt.subplots(2, 2, figsize=(9, 9), sharex=True, sharey=True)
        mus = [[0., 0], [3, -2], [0., 0], [0., 0]]
        sigmas = [np.eye(2), np.eye(2), np.array([[2, 2], [2, 4]]), np.array([[2, -2], [-2, 4]])]
        for mu, sigma, ax in zip(mus, sigmas, axes.ravel()):
            MvNormal(mu, sigma).plot_pdf(marginals=False, ax=ax)

    ========  ==========================
    Support   :math:`x \in \mathbb{R}^k`
    Mean      :math:`\mu`
    Variance  :math:`T^{-1}`
    ========  ==========================

    MvNormal distribution has 2 alternative parameterizations. In terms of the mean and
    the covariance matrix, or in terms of the mean and the precision matrix.

    The link between the 2 alternatives is given by

    .. math::

        T = \Sigma^{-1}

    Parameters
    ----------
    mu : array of floats
        Vector of means.
    cov : array of floats, optional
        Covariance matrix.
    tau : array of floats, optional
        Precision matrix.
    """

    def __init__(self, mu=None, cov=None, tau=None):
        super().__init__()
        self.dist = copy(stats.multivariate_normal)
        self.marginal = Normal
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, cov, tau)

    def _parametrization(self, mu=None, cov=None, tau=None):
        if all_not_none(cov, tau):
            raise ValueError("Incompatible parametrization. Either use mu and cov, or mu and tau.")

        names = ("mu", "cov")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))

        if tau is not None:
            self.tau = tau
            cov = np.linalg.inv(tau)
            names = ("mu", "tau")

        self.mu = mu
        self.cov = cov
        self.param_names = names
        if mu is not None and cov is not None:
            self._update(mu, cov)

    def _get_frozen(self):
        frozen = None
        if all_not_none(self):
            frozen = self.dist(mean=self.mu, cov=self.cov, allow_singular=True)
        return frozen

    def _update(self, mu, cov):
        self.mu = np.array(mu, dtype=float)
        self.cov = np.array(cov, dtype=float)
        self.tau = np.linalg.inv(cov)

        if self.param_names[1] == "cov":
            self.params = (self.mu, self.cov)
        elif self.param_names[1] == "tau":
            self.params = (self.mu, self.tau)

        self._update_rv_frozen()
        self.rv_frozen.var = lambda: np.diag(self.cov)

    def _fit_mle(self, sample, **kwargs):
        raise NotImplementedError

    def plot_pdf(
        self,
        marginals=True,
        pointinterval=False,
        interval=None,
        levels=None,
        support="full",
        legend="title",
        figsize=None,
        ax=None,
    ):
        """
        Plot the pdf of the marginals or the joint pdf
        The joint representation is only available for a 2D Multivariate Normal.

        Parameters
        ----------
        marginals : True
            Defaults to True, plot the marginal distributions, if False plot the joint distribution
            (only valid for a bivariate normal).
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axis
            Axis to plot on

        Returns
        -------
        ax : matplotlib axis
        """
        return plot_mvnormal(
            self, "pdf", marginals, pointinterval, interval, levels, support, legend, figsize, ax
        )

    def plot_cdf(
        self,
        pointinterval=False,
        interval=None,
        levels=None,
        support="full",
        legend="title",
        figsize=None,
        ax=None,
    ):
        """
        Plot the cumulative distribution function.

        Parameters
        ----------
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        support : str:
            If ``full`` use the finite end-points to set the limits of the plot. For unbounded
            end-points or if ``restricted`` use the 0.001 and 0.999 quantiles to set the limits.
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axis
            Axis to plot on

        Returns
        -------
        ax : matplotlib axis
        """
        return plot_mvnormal(
            self, "cdf", "marginals", pointinterval, interval, levels, support, legend, figsize, ax
        )

    def plot_ppf(
        self,
        pointinterval=False,
        interval=None,
        levels=None,
        legend="title",
        figsize=None,
        ax=None,
    ):
        """
        Plot the quantile function.

        Parameters
        ----------
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False. If True the default is to
            plot the median and two interquantiles ranges.
        interval : str
            Type of interval. Available options are highest density interval `"hdi"`,
        equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
        Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        ax : matplotlib axis
            Axis to plot on

        Returns
        -------
        ax : matplotlib axis
        """
        return plot_mvnormal(
            self, "ppf", "marginals", pointinterval, interval, levels, None, legend, figsize, ax
        )

    def plot_interactive(
        self,
        kind="pdf",
        xy_lim="both",
        pointinterval=True,
        interval=None,
        levels=None,
        legend="title",
        figsize=None,
    ):
        """
        Interactive exploration of parameters

        Parameters
        ----------
        kind : str:
            Type of plot. Available options are `pdf`, `cdf` and `ppf`.
        xy_lim : str or tuple
            Set the limits of the x-axis and/or y-axis.
            Defaults to `"both"`, the limits of both axes are fixed for all subplots.
            Use `"auto"` for automatic rescaling of x-axis and y-axis.
            Or set them manually by passing a tuple of 4 elements,
            the first two for x-axis, the last two for y-axis. The tuple can have `None`.
        pointinterval : bool
            Whether to include a plot of the quantiles. Defaults to False.
            If `True` the default is to plot the median and two inter-quantiles ranges.
        interval : str
            Type of interval. Available options are the highest density interval `"hdi"`,
            equal tailed interval `"eti"` or intervals defined by arbitrary `"quantiles"`.
            Defaults to the value in rcParams["stats.ci_kind"].
        levels : list
            Mass of the intervals. For hdi or eti the number of elements should be 2 or 1.
            For quantiles the number of elements should be 5, 3, 1 or 0
            (in this last case nothing will be plotted).
        legend : str
            Whether to include a string with the distribution and its parameter as a ``"title"``
            or not include them ``None``.
        figsize : tuple
            Size of the figure
        """

        check_inside_notebook()

        args = dict(zip(self.param_names, self.params))
        cov, tau = args.get("cov", None), args.get("tau", None)
        self.__init__(**args)  # pylint: disable=unnecessary-dunder-call
        if kind == "pdf":
            w_checkbox_marginals = widgets.Checkbox(
                value=True,
                description="marginals",
                disabled=False,
                indent=False,
            )
            plot_widgets = {"marginals": w_checkbox_marginals}
        else:
            plot_widgets = {}
        for index, mu in enumerate(self.params[0]):
            plot_widgets[f"mu-{index + 1}"] = get_slider(
                f"mu-{index + 1}", mu, *self.params_support[0]
            )

        def plot(**args):
            if kind == "pdf":
                marginals = args.pop("marginals")
            params = {"mu": np.asarray(list(args.values()), dtype=float), "cov": cov, "tau": tau}
            self.__init__(**params)  # pylint: disable=unnecessary-dunder-call
            if kind == "pdf":
                plot_mvnormal(
                    self,
                    "pdf",
                    marginals,
                    pointinterval,
                    interval,
                    levels,
                    "full",
                    legend,
                    figsize,
                    None,
                    xy_lim,
                )

            elif kind == "cdf":
                plot_mvnormal(
                    self,
                    "cdf",
                    "marginals",
                    pointinterval,
                    interval,
                    levels,
                    "full",
                    legend,
                    figsize,
                    None,
                    xy_lim,
                )
            elif kind == "ppf":
                plot_mvnormal(
                    self,
                    "ppf",
                    "marginals",
                    pointinterval,
                    interval,
                    levels,
                    None,
                    legend,
                    figsize,
                    None,
                    xy_lim,
                )

        return interactive(plot, **plot_widgets)



"""Functions to communicate with PPLs."""


from contextlib import contextmanager
import inspect
import logging
import re
import warnings


import matplotlib.pyplot as plt
import numpy as np


from preliz import distributions
from preliz.internal.distribution_helper import init_vals
from preliz.internal.plot_helper import plot_repr
from preliz.distributions import Gamma, Normal, HalfNormal
from preliz.unidimensional.mle import mle
from preliz.ppls.pymc_io import (
    extract_preliz_distributions,
    retrieve_variable_info,
    write_pymc_string,
)
from preliz.ppls.bambi_io import (
    get_pymc_model,
    write_bambi_string,
    dist_as_str,
    match_return_variables,
    dict_model,
)

try:
    from pymc import sample_prior_predictive
except ImportError:
    pass


def posterior_to_prior(model, idata, new_families=None, engine="auto"):
    """
    Fit a posterior from a model to its prior

    The fit is based on maximum likelihood of each posterior marginal to the prior
    in the model. Thus possible correlations between parameters in the posteriors
    will not be preserved. It is expected that the posterior was computed from the model.

    Parameters
    ----------
    model : A PyMC or a Bambi Model
    idata : InferenceData
        InferenceData with a posterior group.
    new_families : "auto", list or dict
        Defaults to None, the samples are fit to the original prior distribution.
        If "auto", the method evaluates the fit to the original prior plus a set of
        predefined distributions.
        Use a list of PreliZ distribution to specify the alternative distributions
        you want to consider.
        Use a dict with variables names in ``model`` as keys and a list of PreliZ
        distributions as values. This allows to specify alternative distributions
        per variable.
    engine : "auto", "pymc" or "bambi"
        Library used to define the model. Either `pymc` or `bambi`. Default is `auto`.
        The function will automatically select the appropriate library to use based on the model
        provided.
    """
    warnings.warn(""""This is an experimental method under development, use with caution.""")
    engine = get_engine(model) if engine == "auto" else engine

    if engine == "bambi":
        model = get_pymc_model(model)

    preliz_model = extract_preliz_distributions(model)
    var_info, _ = retrieve_variable_info(model)

    new_priors = back_fitting_idata(idata, preliz_model, new_families)

    if engine == "bambi":
        new_model = write_bambi_string(new_priors, var_info)
    elif engine == "pymc":
        new_model = write_pymc_string(new_priors, var_info)

    return new_model


def back_fitting_idata(idata, model_info, new_families):
    new_priors = {}
    posterior = idata.posterior.stack(sample=("chain", "draw"))

    if new_families is None:
        for var, dist in model_info.items():
            idx, _ = mle([dist], posterior[var].values, plot=False)
            new_priors[var] = dist
    else:
        for var, dist in model_info.items():
            dists = [dist]

            if new_families == "auto":
                alt = [Normal(), HalfNormal(), Gamma()]
                dists += [a for a in alt if dist.__class__.__name__ != a.__class__.__name__]
            elif isinstance(new_families, list):
                dists += new_families
            elif isinstance(new_families, dict):
                dists += new_families.get(var, [])

            idx, _ = mle(dists, posterior[var].values, plot=False)
            new_priors[var] = dists[idx[0]]
    return new_priors


def inspect_source(fmodel):
    source = inspect.getsource(fmodel)
    signature = inspect.signature(fmodel)
    source = re.sub(r"#.*$|^#.*$", "", source, flags=re.MULTILINE)
    default_params = {
        name: (param.default if param.default is not inspect.Parameter.empty else np.nan)
        for name, param in signature.parameters.items()
    }
    model = fmodel(**default_params)
    return source, signature, get_engine(model)


def get_engine(model):
    if getattr(model, "basic_RVs", False):
        return "pymc"
    elif getattr(model, "formula", False):
        return "bambi"
    return "preliz"


def parse_function_for_pred_textboxes(source, signature, engine="preliz"):
    model = {}

    slidify = list(signature.parameters.keys())
    regex = r"\b" + r"\b|\b".join(slidify) + r"\b"

    all_dist_str = dist_as_str()
    matches = match_preliz_dist(all_dist_str, source, engine)

    for match in matches:
        if engine == "bambi":
            dist_name_str = match.group(1)
        else:
            dist_name_str = match.group(2)

        if engine == "bambi":
            arguments = [s.strip() for s in match.group(2).split(",")]
        else:
            arguments = [s.strip() for s in match.group(3).split(",")]

        args = parse_arguments(arguments, regex, engine)
        for arg in args:
            if arg:
                func, var, idx = arg
                dist = getattr(distributions, dist_name_str)
                model[var] = (dist(**init_vals[dist_name_str]), idx, func)

    return model


def parse_arguments(lst, regex, engine):
    result = []
    if engine == "pymc":
        offset = 1
    else:
        offset = 0
    for idx, item in enumerate(lst):
        match = re.search(regex, item)
        if match:
            if item.isidentifier():
                result.append((None, match.group(0), idx - offset))
            else:
                if "**" in item:
                    power = item.split("**")[1].strip()
                    result.append((power, match.group(0), idx - offset))
                else:
                    func = item.split("(")[0].split(".")[-1]
                    result.append((func, match.group(0), idx - offset))
    return result


def get_prior_pp_samples(fmodel, variables, draws, engine=None, values=None):
    if values is None:
        values = []

    if engine == "preliz":
        obs_rv = variables[-1]  # only one observed for the moment
        pp_samples_ = []
        prior_samples_ = {name: [] for name in variables[:-1]}
        for _ in range(draws):
            for name, value in zip(variables, fmodel(*values)):
                if name == obs_rv:
                    pp_samples_.append(value)
                else:
                    prior_samples_[name].append(value)

        pp_samples = np.stack(pp_samples_)
        prior_samples = {key: np.array(val) for key, val in prior_samples_.items()}
    elif engine == "bambi":
        *prior_samples_, pp_samples = fmodel(*values)
        prior_samples = {name: np.array(val) for name, val in zip(variables[:-1], prior_samples_)}

    return pp_samples, prior_samples


def from_preliz(fmodel):
    source = inspect.getsource(fmodel)
    variables = match_return_variables(source)
    # Find the priors we want to change
    all_dist_str = dist_as_str()
    matches = match_preliz_dist(all_dist_str, source, "preliz")
    # Create a dictionary with the priors
    model = dict_model(matches, variables)

    return variables, model


def match_preliz_dist(all_dist_str, source, engine):
    # remove comments
    source = re.sub(r"#.*$|^#.*$", "", source, flags=re.MULTILINE)

    if engine in ["preliz", "pymc"]:
        regex = rf"(.*?({all_dist_str}).*?)\(([^()]*(?:\([^()]*\)[^()]*)*)\)"
    if engine == "bambi":
        regex = rf'\s*(?:\w+\.)?Prior\("({all_dist_str})",\s*((?:\w+=\w+(?:,?\s*)?)*)\s*\)'
    matches = re.finditer(regex, source)
    return matches


def ppl_plot_decorator(func, iterations, kind_plot, references, plot_func, engine):
    def looper(*args, **kwargs):
        kwargs.pop("__resample__")
        x_min = kwargs.pop("__x_min__")
        x_max = kwargs.pop("__x_max__")
        if not kwargs.pop("__set_xlim__"):
            x_min = None
            x_max = None
            auto = True
        else:
            auto = False

        if engine == "preliz":
            results = []
            for _ in range(iterations):
                val = func(*args, **kwargs)
                if not any(np.isnan(val)):
                    results.append(val)
            results = np.array(results)

        elif engine == "bambi":
            model = func(*args, **kwargs)
            model.build()
            with disable_pymc_sampling_logs():
                idata = model.prior_predictive(iterations)
            results = (
                idata["prior_predictive"]
                .stack(sample=("chain", "draw"))[model.response_component.response.name]
                .values.T
            )

        elif engine == "pymc":
            with func(*args, **kwargs) as model:
                obs_name = model.observed_RVs[0].name
                with disable_pymc_sampling_logs():
                    idata = sample_prior_predictive(samples=iterations)
                results = (
                    idata["prior_predictive"].stack(sample=("chain", "draw"))[obs_name].values.T
                )

        _, ax = plt.subplots()
        ax.set_xlim(x_min, x_max, auto=auto)
        if plot_func is None:
            plot_repr(results, kind_plot, references, iterations, ax)
        else:
            plot_func(results, ax)

    return looper


@contextmanager
def disable_pymc_sampling_logs(logger: logging.Logger = logging.getLogger("pymc")):
    effective_level = logger.getEffectiveLevel()
    logger.setLevel(logging.ERROR)
    try:
        yield
    finally:
        logger.setLevel(effective_level)



import re
from sys import modules
import numpy as np


eps = np.finfo(float).eps


def from_precision(precision):
    sigma = 1 / precision**0.5
    return sigma


def to_precision(sigma):
    precision = 1 / (eps + sigma**2)
    return precision


def all_not_none(*args):
    for arg in args:
        if arg is None:
            return False
    return True


def any_not_none(*args):
    for arg in args:
        if arg is not None:
            return True
    return False


def valid_scalar_params(self, check_frozen=True):
    if not self.is_frozen:
        if check_frozen:
            raise ValueError(
                "Undefined distribution, "
                "you need to first define its parameters or use one of the fit methods"
            )
        return False

    if self.kind not in ["discrete", "continuous"]:
        return True

    if all(
        isinstance(param, (int, float, np.int64)) for param in self.params
    ) or self.__class__.__name__ in ["Categorical", "Mixture"]:
        return True

    raise ValueError("parameters must be integers or floats")


def valid_distribution(self):
    if self.__class__.__name__ != "Categorical":
        return True

    raise ValueError(f"{self.__class__.__name__} is not supported")


def process_extra(input_string):
    pattern = r"(\w+)\((.*?)\)"
    matches = re.findall(pattern, input_string)
    result_dict = {}
    for match in matches:
        name = match[0]
        args = match[1].split(",")
        arg_dict = {}
        try:
            for arg in args:
                key, value = arg.split("=")
                arg_dict[key.strip()] = float(value)
            result_dict[name] = arg_dict
        except ValueError:
            pass
    return result_dict


def num_skewness(dist):
    mean = dist.mean()
    std = dist.std()
    x_values = dist.xvals("full")
    pdf = dist.pdf(x_values)
    if dist.kind == "discrete":
        return np.sum(((x_values - mean) / std) ** 3 * pdf)
    else:
        return np.trapz(((x_values - mean) / std) ** 3 * pdf, x_values)


def num_kurtosis(dist):
    mean = dist.mean()
    std = dist.std()
    x_values = dist.xvals("full")
    pdf = dist.pdf(x_values)
    if dist.kind == "discrete":
        return np.sum(((x_values - mean) / std) ** 4 * pdf) - 3
    else:
        return np.trapz(((x_values - mean) / std) ** 4 * pdf, x_values) - 3


init_vals = {
    "AsymmetricLaplace": {"kappa": 1.0, "mu": 0.0, "b": 1.0},
    "Beta": {"alpha": 2, "beta": 2},
    "BetaScaled": {"alpha": 2, "beta": 2, "lower": -1.0, "upper": 2.0},
    "Cauchy": {"alpha": 0.0, "beta": 1.0},
    "ChiSquared": {"nu": 5.0},
    "ExGaussian": {"mu": 0.0, "sigma": 1, "nu": 2},
    "Exponential": {"lam": 1},
    "Gamma": {"alpha": 2.0, "beta": 5.0},
    "Gumbel": {"mu": 2.0, "beta": 5.0},
    "HalfCauchy": {"beta": 1.0},
    "HalfNormal": {"sigma": 1.0},
    "HalfStudentT": {"nu": 7.0, "sigma": 1.0},
    "InverseGamma": {"alpha": 3.0, "beta": 5.0},
    "Kumaraswamy": {"a": 2.0, "b": 2.0},
    "Laplace": {"mu": 0.0, "b": 1.0},
    "LogLogistic": {"alpha": 1, "beta": 5},
    "Logistic": {"mu": 0.0, "s": 1},
    "LogNormal": {"mu": 0.0, "sigma": 0.5},
    "LogitNormal": {"mu": 0.0, "sigma": 0.5},
    "Moyal": {"mu": 0.0, "sigma": 1.0},
    "Normal": {"mu": 0.0, "sigma": 1.0},
    "Pareto": {"alpha": 5, "m": 2.0},
    "Rice": {"nu": 2.0, "sigma": 1.0},
    "SkewNormal": {"mu": 0.0, "sigma": 1, "alpha": 6.0},
    "SkewStudentT": {"mu": 0.0, "sigma": 1, "a": 2, "b": 2},
    "StudentT": {"nu": 7, "mu": 0.0, "sigma": 1},
    "Triangular": {"lower": -2, "c": 0.0, "upper": 2.0},
    "TruncatedNormal": {"mu": 0.0, "sigma": 1, "lower": -2, "upper": 3.0},
    "Uniform": {"lower": 0.0, "upper": 1.0},
    "VonMises": {"mu": 0.0, "kappa": 1.0},
    "Wald": {"mu": 1, "lam": 3.0},
    "Weibull": {"alpha": 5.0, "beta": 2.0},
    "Bernoulli": {"p": 0.5},
    "BetaBinomial": {"alpha": 2, "beta": 2, "n": 10},
    "Binomial": {"n": 5, "p": 0.5},
    "Categorical": {"p": [0.5, 0.1, 0.4]},
    "DiscreteUniform": {"lower": -2.0, "upper": 2.0},
    "DiscreteWeibull": {"q": 0.9, "beta": 1.3},
    "Geometric": {"p": 0.5},
    "HyperGeometric": {"N": 50, "k": 10, "n": 20},
    "NegativeBinomial": {"mu": 5.0, "alpha": 2.0},
    "Poisson": {"mu": 4.5},
    "ZeroInflatedBinomial": {"psi": 0.7, "n": 10, "p": 0.5},
    "ZeroInflatedNegativeBinomial": {"psi": 0.7, "mu": 5, "alpha": 8},
    "ZeroInflatedPoisson": {"psi": 0.8, "mu": 4.5},
    "Truncated": {"lower": -10, "upper": 10},
    "Censored": {"lower": -10, "upper": 10},
}


def get_distributions(dist_names=None):

    if dist_names is None:
        all_distributions = modules["preliz.distributions"].__all__
    else:
        all_distributions = dist_names

    distributions = []
    for a_dist in all_distributions:
        dist = getattr(modules["preliz.distributions"], a_dist)()
        distributions.append(dist)
    return distributions



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, to_precision, from_precision, all_not_none
from ..internal.special import erf, erfinv, logit, expit, mean_and_std, cdf_bounds, ppf_bounds_cont


class LogitNormal(Continuous):
    r"""
    Logit-Normal distribution.

    The pdf of this distribution is

    .. math::
       f(x \mid \mu, \tau) =
           \frac{1}{x(1-x)} \sqrt{\frac{\tau}{2\pi}}
           \exp\left\{ -\frac{\tau}{2} (logit(x)-\mu)^2 \right\}


    .. plot::
        :context: close-figs


        from preliz import LogitNormal, style
        style.use('preliz-doc')
        mus = [0., 0., 0., 1.]
        sigmas = [0.3, 1., 2., 1.]
        for mu, sigma in zip(mus, sigmas):
            LogitNormal(mu, sigma).plot_pdf()

    ========  ==========================================
    Support   :math:`x \in (0, 1)`
    Mean      no analytical solution
    Variance  no analytical solution
    ========  ==========================================

    Parameters
    ----------
    mu : float
        Location parameter.
    sigma : float
        Scale parameter (sigma > 0).
    tau : float
        Scale parameter (tau > 0).
    """

    def __init__(self, mu=None, sigma=None, tau=None):
        super().__init__()
        self.support = (0, 1)
        self._parametrization(mu, sigma, tau)

    def _parametrization(self, mu=None, sigma=None, tau=None):
        if all_not_none(sigma, tau):
            raise ValueError(
                "Incompatible parametrization. Either use mu and sigma, or mu and tau."
            )

        names = ("mu", "sigma")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))

        if tau is not None:
            self.tau = tau
            sigma = from_precision(tau)
            names = ("mu", "tau")

        self.mu = mu
        self.sigma = sigma
        self.param_names = names
        if all_not_none(mu, sigma):
            self._update(mu, sigma)

    def _get_frozen(self):
        frozen = None
        if all_not_none(self.params):
            frozen = self.dist(self.mu, self.sigma)
        return frozen

    def _update(self, mu, sigma):
        self.mu = np.float64(mu)
        self.sigma = np.float64(sigma)
        self.tau = to_precision(sigma)

        if self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma)
        elif self.param_names[1] == "tau":
            self.params = (self.mu, self.tau)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.sigma)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.sigma)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.sigma)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.sigma)

    def entropy(self):
        x_values = self.xvals("restricted")
        logpdf = self.logpdf(x_values)
        return -np.trapz(np.exp(logpdf) * logpdf, x_values)

    def mean(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.trapz(x_values * pdf, x_values)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.trapz((x_values - self.mean()) ** 2 * pdf, x_values)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        mean = self.mean()
        std = self.std()
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.trapz(((x_values - mean) / std) ** 3 * pdf, x_values)

    def kurtosis(self):
        mean = self.mean()
        std = self.std()
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.trapz(((x_values - mean) / std) ** 4 * pdf, x_values) - 3

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return expit(random_state.normal(self.mu, self.sigma, size))

    def _fit_moments(self, mean, sigma):
        mu = logit(mean)
        sigma = np.diff((mean - sigma * 3, mean + sigma * 3))
        self._update(mu, sigma)

    def _fit_mle(self, sample):
        mu, sigma = mean_and_std(logit(sample))
        self._update(mu, sigma)


@nb.njit(cache=True)
def nb_cdf(x, mu, sigma):
    return cdf_bounds(0.5 * (1 + erf((logit(x) - mu) / (sigma * 2**0.5))), x, 0, 1)


@nb.njit(cache=True)
def nb_ppf(q, mu, sigma):
    return ppf_bounds_cont(expit(mu + sigma * 2**0.5 * erfinv(2 * q - 1)), q, 0, 1)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, sigma):
    if x <= 0:
        return -np.inf
    if x >= 1:
        return -np.inf
    else:
        return (
            -np.log(sigma)
            - 0.5 * np.log(2 * np.pi)
            - 0.5 * ((logit(x) - mu) / sigma) ** 2
            - np.log(x)
            - np.log1p(-x)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, sigma):
    return -(nb_logpdf(x, mu, sigma)).sum()



# Continuous Distributions
from .asymmetric_laplace import AsymmetricLaplace
from .beta import Beta
from .betascaled import BetaScaled
from .cauchy import Cauchy
from .chi_squared import ChiSquared
from .exgaussian import ExGaussian
from .exponential import Exponential
from .gamma import Gamma
from .gumbel import Gumbel
from .halfcauchy import HalfCauchy
from .halfnormal import HalfNormal
from .halfstudentt import HalfStudentT
from .inversegamma import InverseGamma
from .kumaraswamy import Kumaraswamy
from .laplace import Laplace
from .logistic import Logistic
from .loglogistic import LogLogistic
from .logitnormal import LogitNormal
from .lognormal import LogNormal
from .moyal import Moyal
from .normal import Normal
from .pareto import Pareto
from .skew_studentt import SkewStudentT
from .skewnormal import SkewNormal
from .studentt import StudentT
from .rice import Rice
from .triangular import Triangular
from .truncatednormal import TruncatedNormal
from .uniform import Uniform
from .vonmises import VonMises
from .wald import Wald
from .weibull import Weibull

# Multivariate Distributions
from .continuous_multivariate import *

# Discrete Distributions
from .bernoulli import Bernoulli
from .betabinomial import BetaBinomial
from .binomial import Binomial
from .categorical import Categorical
from .discrete_uniform import DiscreteUniform
from .discrete_weibull import DiscreteWeibull
from .geometric import Geometric
from .hypergeometric import HyperGeometric
from .poisson import Poisson
from .negativebinomial import NegativeBinomial
from .zi_binomial import ZeroInflatedBinomial
from .zi_negativebinomial import ZeroInflatedNegativeBinomial
from .zi_poisson import ZeroInflatedPoisson

# Transform Distributions
from .mixture import Mixture
from .truncated import Truncated
from .censored import Censored
from .hurdle import Hurdle


all_continuous = [
    AsymmetricLaplace,
    Beta,
    BetaScaled,
    Cauchy,
    ChiSquared,
    ExGaussian,
    Exponential,
    Gamma,
    Gumbel,
    HalfCauchy,
    HalfNormal,
    HalfStudentT,
    InverseGamma,
    Kumaraswamy,
    Laplace,
    Logistic,
    LogLogistic,
    LogNormal,
    LogitNormal,
    Moyal,
    Normal,
    Pareto,
    Rice,
    SkewNormal,
    SkewStudentT,
    StudentT,
    Triangular,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
]
all_discrete = [
    Bernoulli,
    BetaBinomial,
    Binomial,
    Categorical,
    DiscreteUniform,
    DiscreteWeibull,
    Geometric,
    HyperGeometric,
    NegativeBinomial,
    Poisson,
    ZeroInflatedBinomial,
    ZeroInflatedNegativeBinomial,
    ZeroInflatedPoisson,
]

all_continuous_multivariate = [Dirichlet, MvNormal]


__all__ = (
    [s.__name__ for s in all_continuous]
    + [s.__name__ for s in all_discrete]
    + [s.__name__ for s in all_continuous_multivariate]
    + [Mixture.__name__]
    + [Truncated.__name__]
    + [Censored.__name__]
    + [Hurdle.__name__]
)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from scipy.special import gammainc, gammaincinv  # pylint: disable=no-name-in-module

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.special import cdf_bounds, ppf_bounds_cont, gammaln, digamma, xlogy
from ..internal.optimization import optimize_ml


class ChiSquared(Continuous):
    r"""
    Chi squared  distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \nu) =
                \frac{x^{(\nu-2)/2}e^{-x/2}}{2^{\nu/2}\Gamma(\nu/2)}

    .. plot::
        :context: close-figs


        from preliz import ChiSquared, style
        style.use('preliz-doc')
        nus = [1., 3., 9.]
        for nu in nus:
                ax = ChiSquared(nu).plot_pdf(support=(0,20))
                ax.set_ylim(0, 0.6)

    ========  ===============================
    Support   :math:`x \in [0, \infty)`
    Mean      :math:`\nu`
    Variance  :math:`2 \nu`
    ========  ===============================

    Parameters
    ----------
    nu : float
        Degrees of freedom (nu > 0).
    """

    def __init__(self, nu=None):
        super().__init__()
        self.nu = nu
        self.support = (0, np.inf)
        self._parametrization(nu)

    def _parametrization(self, nu=None):
        self.nu = nu
        self.param_names = ("nu",)
        self.params_support = ((eps, np.inf),)
        self.params = (self.nu,)
        if self.nu is not None:
            self._update(self.nu)

    def _get_frozen(self):
        frozen = None
        if all_not_none(self.params):
            frozen = self.dist(self.nu)
        return frozen

    def _update(self, nu):
        self.nu = np.float64(nu)
        self.params = (self.nu,)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.nu))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.nu)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.nu)

    def logpdf(self, x):
        return nb_logpdf(x, self.nu)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.nu)

    def entropy(self):
        return nb_entropy(self.nu)

    def mean(self):
        return self.nu

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.nu * 2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (8 / self.nu) ** 0.5

    def kurtosis(self):
        return 12 / self.nu

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.chisquare(self.nu, size)

    def _fit_moments(self, mean, sigma=None):  # pylint: disable=unused-argument
        self._update(mean)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


# @nb.njit(cache=True)
def nb_cdf(x, nu):
    return cdf_bounds(gammainc(nu / 2, x / 2), x, 0, np.inf)


# @nb.njit(cache=True)
def nb_ppf(q, nu):
    vals = 2 * gammaincinv(nu / 2, q)
    return ppf_bounds_cont(vals, q, 0, np.inf)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, nu):
    if x < 0:
        return -np.inf
    else:
        return xlogy(nu / 2 - 1, x) - x / 2 - gammaln(nu / 2) - (nu * np.log(2)) / 2


@nb.njit(cache=True)
def nb_neg_logpdf(x, lam):
    return (-nb_logpdf(x, lam)).sum()


@nb.njit(cache=True)
def nb_entropy(nu):
    h_nu = nu / 2
    return h_nu + np.log(2) + gammaln(h_nu) + (1 - h_nu) * digamma(h_nu)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np
from scipy.special import bdtr, bdtrik  # pylint: disable=no-name-in-module

from .distributions import Discrete
from ..internal.optimization import optimize_moments
from ..internal.distribution_helper import eps, all_not_none
from ..internal.special import cdf_bounds, ppf_bounds_disc, gammaln, mean_and_std, xlogy, xlog1py


class Binomial(Discrete):
    R"""
    Binomial distribution.

    The discrete probability distribution of the number of successes
    in a sequence of n independent yes/no experiments, each of which
    yields success with probability p.

    The pmf of this distribution is

    .. math:: f(x \mid n, p) = \binom{n}{x} p^x (1-p)^{n-x}

    .. plot::
        :context: close-figs


        from preliz import Binomial, style
        style.use('preliz-doc')
        ns = [5, 10, 10]
        ps = [0.5, 0.5, 0.7]
        for n, p in zip(ns, ps):
            Binomial(n, p).plot_pdf()

    ========  ==========================================
    Support   :math:`x \in \{0, 1, \ldots, n\}`
    Mean      :math:`n p`
    Variance  :math:`n p (1 - p)`
    ========  ==========================================

    Parameters
    ----------
    n : int
        Number of Bernoulli trials (n >= 0).
    p : float
        Probability of success in each trial (0 < p < 1).
    """

    def __init__(self, n=None, p=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(n, p)

    def _parametrization(self, n=None, p=None):
        self.n = n
        self.p = p
        self.params = (self.n, self.p)
        self.param_names = ("n", "p")
        self.params_support = ((eps, np.inf), (eps, 1 - eps))
        if all_not_none(n, p):
            self._update(n, p)

    def _update(self, n, p):
        self.n = np.int64(n)
        self.p = np.float64(p)
        self._q = 1 - self.p
        self.params = (self.n, self.p)
        self.support = (0, self.n)
        self.is_frozen = True

    def _fit_moments(self, mean, sigma):
        # crude approximation for n and p
        n = mean + sigma * 2
        p = mean / n
        params = n, p
        optimize_moments(self, mean, sigma, params)

    def _fit_mle(self, sample):
        self._update(*nb_fit_mle(sample))

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        return nb_cdf(x, self.n, self.p, self.support[0], self.support[1])

    def ppf(self, q):
        return nb_ppf(q, self.n, self.p, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.n, self.p)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.n, self.p)

    def entropy(self):
        return nb_entropy(self.n, self.p)

    def mean(self):
        return self.n * self.p

    def median(self):
        return np.ceil(self.n * self.p)

    def var(self):
        return self.n * self.p * self._q

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return (self._q - self.p) / self.std()

    def kurtosis(self):
        return (1 - 6 * self.p * self._q) / (self.n * self.p * self._q)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.binomial(self.n, self.p, size=size)

    def _fit_moments(self, mean, sigma):
        # crude approximation for n and p
        n = mean + sigma * 2
        p = mean / n
        params = n, p
        return optimize_moments(self, mean, sigma, params)

    def _fit_mle(self, sample):
        self._update(*nb_fit_mle(sample))


# @nb.jit
# bdtr not supported by numba
def nb_cdf(x, n, p, lower, upper):
    x = np.asarray(x)
    prob = np.asarray(bdtr(x, n, p))
    return cdf_bounds(prob, x, lower, upper)


# @nb.jit
def nb_ppf(q, n, p, lower, upper):
    q = np.asarray(q)
    x_vals = np.ceil(bdtrik(q, n, p))
    return ppf_bounds_disc(x_vals, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(n, p):
    return 0.5 * np.log(2 * np.pi * np.e * n * p * (1 - p))


@nb.njit(cache=True)
def nb_fit_mle(sample):
    # see https://doi.org/10.1016/j.jspi.2004.02.019 for details
    x_bar, x_std = mean_and_std(sample)
    x_max = np.max(sample)
    n = np.ceil(x_max ** (1.5) * x_std / (x_bar**0.5 * (x_max - x_bar) ** 0.5))
    p = x_bar / n
    return n, p


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, n, p):
    if x < 0:
        return -np.inf
    elif x > n:
        return -np.inf
    else:
        return (
            gammaln(n + 1)
            - (gammaln(x + 1) + gammaln(n - x + 1))
            + xlogy(x, p)
            + xlog1py(n - x, -p)
        )


@nb.njit(cache=True)
def nb_neg_logpdf(x, n, p):
    return -(nb_logpdf(x, n, p)).sum()



"""Projective predictive elicitation."""

import warnings
import numpy as np

from preliz.internal.optimization import optimize_pymc_model
from preliz.ppls.bambi_io import get_pymc_model, write_bambi_string
from preliz.ppls.agnostic import back_fitting_idata, get_engine
from preliz.ppls.pymc_io import (
    extract_preliz_distributions,
    retrieve_variable_info,
    unravel_projection,
    get_initial_guess,
    compile_mllk,
    back_fitting_pymc,
    write_pymc_string,
)


def ppe(model, target, method="projective", engine="auto", random_state=0):
    """
    Prior Predictive Elicitation.

    This method is experimental and under development. It does not offers guarantees of
    correctness. Use with caution and triple-check the results.

    Parameters
    ----------
    model : a probabilistic model
        Currently it only works with PyMC model. More PPls coming soon.
    method : str
        Method used to generate samples that match the target distribution.
        Defaults to `"projective"`, another option is `"pathfinder"`.
        If `"projective"`, the parameters of the priors are only used to provide an initial
        guess for the optimization routine. Thus their effect on the result is smaller than in
        traditional Bayesian inference, unless the priors are very vague or very strong.
        If `"projective"`, the observed values are ignored, but not their size.
        Pathfinder is a variational inference method so the role of the priors and observed values
        is what is expected in Bayesian inference.
    engine : str
        Library used to define the model. Either `"auto"` (default), `"pymc"` or `"bambi"`.
        Ig `"auto"`, the library is automatically detected.
    target : a PreliZ distribution or list
        Instance of a PreliZ distribution or a list of tuples where each tuple contains a PreliZ
        distribution and a weight.
        This represents the prior predictive distribution **previously** elicited by the user,
        possibly using other PreliZ's methods to obtain this distribution, such as maxent,
        roulette, quartile, etc.
        This should represent the domain-knowledge of the user and not any observed dataset.
    random_state : {None, int, numpy.random.Generator, numpy.random.RandomState}
        Defaults to 0. Ignored if `method` is `"pathfinder"`.

    Returns
    -------
    new_priors : str
        A string representation of the new priors. The user can copy and paste it into
        the model's code. Ideally, with none to minimal changes.
    """
    warnings.warn(
        """This method is experimental and under development with no guarantees of correctness.
                  Use with caution and triple-check the results."""
    )
    opt_iterations = 400

    rng = np.random.default_rng(random_state)
    engine = get_engine(model) if engine == "auto" else engine

    # Get models information
    if engine == "bambi":
        model = get_pymc_model(model)

    preliz_model = extract_preliz_distributions(model)
    var_info, num_draws = retrieve_variable_info(model)

    # With the projective method we attempt to find a prior that induces
    # a prior predictive distribution as close as possible to the target distribution
    if method == "projective":
        # Initial point for optimization
        initial_guess = get_initial_guess(model)
        # compile PyMC model
        fmodel, old_y_value, obs_rvs = compile_mllk(model)
        projection_raveled = optimize_pymc_model(
            fmodel,
            target,
            num_draws,
            opt_iterations,
            initial_guess,
            rng,
        )
        # restore obs_rvs value in the model
        model.rvs_to_values[obs_rvs] = old_y_value

        projection_unraveled = unravel_projection(projection_raveled, var_info, opt_iterations)

        # Backfit `projected_posterior` into the model's prior-families
        projection_backfitted = back_fitting_pymc(projection_unraveled, preliz_model, var_info)

        if engine == "bambi":
            new_priors = write_bambi_string(projection_backfitted, var_info)
        elif engine == "pymc":
            new_priors = write_pymc_string(projection_backfitted, var_info)

    elif method == "pathfinder":
        from pymc_experimental import fit  # pylint:disable=import-outside-toplevel

        with model:
            idata = fit(method="pathfinder", num_samples=opt_iterations)

        projection_backfitted = back_fitting_idata(idata, preliz_model, new_families=False)
        if engine == "bambi":
            new_priors = write_bambi_string(projection_backfitted, var_info)
        elif engine == "pymc":
            new_priors = write_pymc_string(projection_backfitted, var_info)

    return new_priors



"""Functions originally imported from ArviZ"""
import warnings

import numpy as np
from numba import njit

from scipy.fftpack import fft
from scipy.signal import convolve
from scipy.signal.windows import gaussian

from .optimization import _root
from .rcparams import rcParams


def hdi(ary, hdi_prob=None, skipna=True):
    """
    Calculate highest density interval (HDI) of array for given probability.

    The HDI is the minimum width Bayesian credible interval (BCI).

    Parameters
    ----------
    ary: array_like
        An array containing the values for which the HDI is to be computed.
    hdi_prob: float, optional
        Prob for which the highest density interval will be computed. Defaults to None,
        which results in the value of rcParams["stats.ci_prob"] being used.

    Returns
    -------
    np.ndarray with lower and upper values of the interval.
    """
    if hdi_prob is None:
        hdi_prob = rcParams["stats.ci_prob"]

    if not 1 >= hdi_prob > 0:
        raise ValueError("The value of hdi_prob should be in the interval (0, 1]")

    ary = np.ravel(ary)
    if skipna:
        nans = np.isnan(ary)
        if not nans.all():
            ary = ary[~nans]
    n = len(ary)

    ary = np.sort(ary)
    interval_idx_inc = int(np.floor(hdi_prob * n))
    n_intervals = n - interval_idx_inc
    interval_width = np.subtract(ary[interval_idx_inc:], ary[:n_intervals], dtype=np.float64)

    if len(interval_width) == 0:
        raise ValueError("Too few elements for interval calculation. ")

    min_idx = np.argmin(interval_width)
    hdi_min = ary[min_idx]
    hdi_max = ary[min_idx + interval_idx_inc]
    hdi_interval = np.array([hdi_min, hdi_max])

    return hdi_interval


def kde(x):
    x = x[np.isfinite(x)]
    if x.size == 0 or np.all(x == x[0]):
        warnings.warn("Your data appears to have a single value or no finite values")

        return np.zeros(2), np.array([np.nan] * 2)

    grid_len = 256
    # Preliminary calculations
    x_min = x.min()
    x_max = x.max()
    x_range = x_max - x_min

    # Determine grid
    grid_min = x_min
    grid_max = x_max

    grid_counts, _, grid_edges = histogram(x, grid_len, (grid_min, grid_max))

    # Bandwidth estimation

    band_w = _bw_isj(x, grid_counts=grid_counts, x_range=x_range)

    # Density estimation
    grid, pdf = _kde_convolution(x, band_w, grid_edges, grid_counts, grid_len)

    return grid, pdf


@njit(cache=True)
def histogram(data, bins, range_hist=None):
    """Jitted histogram.

    Parameters
    ----------
    data : array-like
        Input data. Passed as first positional argument to ``np.histogram``.
    bins : int or array-like
        Passed as keyword argument ``bins`` to ``np.histogram``.
    range_hist : (float, float), optional
        Passed as keyword argument ``range`` to ``np.histogram``.

    Returns
    -------
    hist : array
        The number of counts per bin.
    density : array
        The density corresponding to each bin.
    bin_edges : array
        The edges of the bins used.
    """
    hist, bin_edges = np.histogram(data, bins=bins, range=range_hist)
    hist_dens = hist / (hist.sum() * np.diff(bin_edges))
    return hist, hist_dens, bin_edges


def _kde_convolution(x, band_w, grid_edges, grid_counts, grid_len):
    """Kernel density with convolution.

    One dimensional Gaussian kernel density estimation via convolution of the binned relative
    frequencies and a Gaussian filter. This is an internal function used by `kde()`.
    """
    # Calculate relative frequencies per bin
    bin_width = grid_edges[1] - grid_edges[0]
    freq = grid_counts / bin_width / len(x)

    # Bandwidth must consider the bin width
    band_w /= bin_width

    grid = (grid_edges[1:] + grid_edges[:-1]) / 2

    kernel_n = int(band_w * 2 * np.pi)
    if kernel_n == 0:
        kernel_n = 1

    kernel = gaussian(kernel_n, band_w)

    npad = int(grid_len / 5)
    freq = np.concatenate([freq[npad - 1 :: -1], freq, freq[grid_len : grid_len - npad - 1 : -1]])
    pdf = convolve(freq, kernel, mode="same", method="direct")[npad : npad + grid_len]
    pdf /= band_w * (2 * np.pi) ** 0.5

    return grid, pdf


def _bw_isj(x, grid_counts=None, x_range=None):
    """Improved Sheather-Jones bandwidth estimation.

    Improved Sheather and Jones method as explained in [1]_. This method is used internally by the
    KDE estimator, resulting in saved computation time as minimums, maximums and the grid are
    pre-computed.

    References
    ----------
    .. [1] Kernel density estimation via diffusion.
       Z. I. Botev, J. F. Grotowski, and D. P. Kroese.
       Ann. Statist. 38 (2010), no. 5, 2916--2957.
    """
    x_len = len(x)
    grid_len = len(grid_counts) - 1
    # Discrete cosine transform of the data
    a_k = _dct1d(grid_counts / x_len)

    k_sq = np.arange(1, grid_len) ** 2
    a_sq = a_k[range(1, grid_len)] ** 2
    return _root(x_len, k_sq, a_sq, x) ** 0.5 * x_range


def _dct1d(x):
    """Discrete Cosine Transform in 1 Dimension.

    Parameters
    ----------
    x : numpy array
        1 dimensional array of values for which the
        DCT is desired

    Returns
    -------
    output : DTC transformed values
    """
    x_len = len(x)

    even_increasing = np.arange(0, x_len, 2)
    odd_decreasing = np.arange(x_len - 1, 0, -2)

    x = np.concatenate((x[even_increasing], x[odd_decreasing]))

    w_1k = np.r_[1, (2 * np.exp(-(0 + 1j) * (np.arange(1, x_len)) * np.pi / (2 * x_len)))]
    output = np.real(w_1k * fft(x))

    return output



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, any_not_none, all_not_none
from ..internal.optimization import optimize_ml
from ..internal.special import (
    betaln,
    betainc,
    betaincinv,
    digamma,
    gammaln,
    cdf_bounds,
    ppf_bounds_cont,
    mean_and_std,
    xlogy,
    xlog1py,
)


class Beta(Continuous):
    r"""
    Beta distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \alpha, \beta) =
           \frac{x^{\alpha - 1} (1 - x)^{\beta - 1}}{B(\alpha, \beta)}

    .. plot::
        :context: close-figs

        
        from preliz import Beta, style
        style.use('preliz-doc')
        alphas = [.5, 5., 2.]
        betas = [.5, 5., 5.]
        for alpha, beta in zip(alphas, betas):
            ax = Beta(alpha, beta).plot_pdf()
        ax.set_ylim(0, 5)

    ========  ==============================================================
    Support   :math:`x \in (0, 1)`
    Mean      :math:`\dfrac{\alpha}{\alpha + \beta}`
    Variance  :math:`\dfrac{\alpha \beta}{(\alpha+\beta)^2(\alpha+\beta+1)}`
    ========  ==============================================================

    Beta distribution has 3 alternative parameterizations. In terms of alpha and
    beta, mean and sigma (standard deviation) or mean and nu (concentration).

    The link between the 3 alternatives is given by

    .. math::

       \alpha &= \mu \nu \\
       \beta  &= (1 - \mu) \nu

       \text{where } \nu = \frac{\mu(1-\mu)}{\sigma^2} - 1


    Parameters
    ----------
    alpha : float
        alpha  > 0
    beta : float
        beta  > 0
    mu : float
        mean (0 < ``mu`` < 1).
    sigma : float
        standard deviation (``sigma`` < sqrt(``mu`` * (1 - ``mu``))).
    nu : float
        concentration > 0
    """

    def __init__(self, alpha=None, beta=None, mu=None, sigma=None, nu=None):
        super().__init__()
        self.support = (0, 1)
        self._parametrization(alpha, beta, mu, sigma, nu)

    def _parametrization(self, alpha=None, beta=None, mu=None, sigma=None, nu=None):
        if any_not_none(alpha, beta) and any_not_none(mu, sigma, nu) or all_not_none(sigma, nu):
            raise ValueError(
                "Incompatible parametrization. Either use alpha and beta, or mu and sigma."
            )

        self.param_names = ("alpha", "beta")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if any_not_none(mu, sigma):
            self.mu = mu
            self.sigma = sigma
            self.param_names = ("mu", "sigma")
            self.params_support = ((eps, 1 - eps), (eps, 1 - eps))
            if all_not_none(mu, sigma):
                alpha, beta = self._from_mu_sigma(mu, sigma)

        if any_not_none(mu, nu) and sigma is None:
            self.mu = mu
            self.nu = nu
            self.param_names = ("mu", "nu")
            self.params_support = ((eps, 1 - eps), (eps, np.inf))
            if all_not_none(mu, nu):
                alpha, beta = self._from_mu_nu(mu, nu)

        self.alpha = alpha
        self.beta = beta
        if all_not_none(self.alpha, self.beta):
            self._update(self.alpha, self.beta)

    def _from_mu_sigma(self, mu, sigma):
        nu = mu * (1 - mu) / sigma**2 - 1
        alpha = mu * nu
        beta = (1 - mu) * nu
        return alpha, beta

    def _from_mu_nu(self, mu, nu):
        alpha = mu * nu
        beta = (1 - mu) * nu
        return alpha, beta

    def _to_mu_sigma(self, alpha, beta):
        alpha_plus_beta = alpha + beta
        mu = alpha / alpha_plus_beta
        sigma = (alpha * beta) ** 0.5 / alpha_plus_beta / (alpha_plus_beta + 1) ** 0.5
        return mu, sigma

    def _update(self, alpha, beta):
        self.alpha = np.float64(alpha)
        self.beta = np.float64(beta)
        self.mu, self.sigma = self._to_mu_sigma(self.alpha, self.beta)
        self.nu = self.mu * (1 - self.mu) / self.sigma**2 - 1

        if self.param_names[0] == "alpha":
            self.params = (self.alpha, self.beta)
        elif self.param_names[1] == "sigma":
            self.params = (self.mu, self.sigma)
        elif self.param_names[1] == "nu":
            self.params = (self.mu, self.nu)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.alpha, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.beta, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.beta, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.beta)

    def entropy(self):
        return nb_entropy(self.alpha, self.beta)

    def mean(self):
        return self.alpha / (self.alpha + self.beta)

    def mode(self):
        return np.where(
            (self.alpha > 1) & (self.beta > 1), (self.alpha - 1) / (self.alpha + self.beta - 2), 0.5
        )

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return (self.alpha * self.beta) / (
            (self.alpha + self.beta) ** 2 * (self.alpha + self.beta + 1)
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        if self.alpha == self.beta:
            return np.zeros_like(self.alpha)
        else:
            psc = self.alpha + self.beta
            return (2 * (self.beta - self.alpha) * np.sqrt(psc + 1)) / (
                (psc + 2) * np.sqrt(self.alpha * self.beta)
            )

    def kurtosis(self):
        psc = self.alpha + self.beta
        prod = self.alpha * self.beta
        return (
            6
            * (np.abs(self.alpha - self.beta) ** 2 * (psc + 1) - prod * (psc + 2))
            / (prod * (psc + 2) * (psc + 3))
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.beta(self.alpha, self.beta, size)

    def _fit_moments(self, mean, sigma):
        alpha, beta = self._from_mu_sigma(mean, sigma)
        alpha = max(0.5, alpha)
        beta = max(0.5, beta)
        self._update(alpha, beta)

    def _fit_mle(self, sample):
        mean, std = mean_and_std(sample)
        self._fit_moments(mean, std)
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, alpha, beta, lower, upper):
    prob = betainc(alpha, beta, x)
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, alpha, beta, lower, upper):
    x_val = betaincinv(alpha, beta, q)
    return ppf_bounds_cont(x_val, q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(alpha, beta):
    psc = alpha + beta
    return (
        betaln(alpha, beta)
        - (alpha - 1) * digamma(alpha)
        - (beta - 1) * digamma(beta)
        + (psc - 2) * digamma(psc)
    )


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, beta):
    if x <= 0 or x >= 1:
        return -np.inf
    else:
        beta_ = gammaln(alpha) + gammaln(beta) - gammaln(alpha + beta)
        return xlogy((alpha - 1), x) + xlog1py((beta - 1), -x) - beta_


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, beta):
    return -(nb_logpdf(x, alpha, beta)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.optimization import optimize_ml, optimize_moments
from ..internal.special import (
    beta,
    digamma,
    cdf_bounds,
    ppf_bounds_cont,
    xlogy,
    xlog1py,
)


class Kumaraswamy(Continuous):
    r"""
    Kumaraswamy distribution.

    The pdf of this distribution is

    .. math::

         f(x \mid a, b) = a b x^{a - 1} (1 - x^a)^{b - 1}

    .. plot::
        :context: close-figs


        from preliz import Kumaraswamy, style
        style.use('preliz-doc')
        a_s = [.5, 5., 1., 2., 2.]
        b_s = [.5, 1., 3., 2., 5.]
        for a, b in zip(a_s, b_s):
            ax = Kumaraswamy(a, b).plot_pdf()
            ax.set_ylim(0, 3.)

    ========  ==============================================================
    Support   :math:`x \in (0, 1)`
    Mean      :math:`b B(1 + \tfrac{1}{a}, b)`
    Variance  :math:`b B(1 + \tfrac{2}{a}, b) - (b B(1 + \tfrac{1}{a}, b))^2`
    ========  ==============================================================

    Parameters
    ----------
    a : float
        a > 0.
    b : float
        b > 0.
    """

    def __init__(self, a=None, b=None):
        super().__init__()
        self.support = (0, 1)
        self._parametrization(a, b)

    def _parametrization(self, a=None, b=None):
        self.a = a
        self.b = b
        self.params = (self.a, self.b)
        self.param_names = ("a", "b")
        self.params_support = ((eps, np.inf), (eps, np.inf))
        if (a and b) is not None:
            self._update(a, b)

    def _get_frozen(self):
        frozen = None
        if all_not_none(self.params):
            frozen = self.dist(self.a, self.b)
        return frozen

    def _update(self, a, b):
        self.a = np.float64(a)
        self.b = np.float64(b)
        self.params = (self.a, self.b)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.a, self.b))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.a, self.b)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.a, self.b)

    def logpdf(self, x):
        return nb_logpdf(x, self.a, self.b)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.a, self.b)

    def entropy(self):
        return nb_entropy(self.a, self.b)

    def mean(self):
        return _mom(self.a, self.b, 1)

    def median(self):
        return (1 - 2 ** -(1 / self.b)) ** (1 / self.a)

    def var(self):
        m_1 = _mom(self.a, self.b, 1)
        m_2 = _mom(self.a, self.b, 2)
        return m_2 - m_1**2

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        mean = self.mean()
        var = self.var()
        m_3 = _mom(self.a, self.b, 3)
        return (m_3 - 3 * mean * var - mean**3) / var**1.5

    def kurtosis(self):
        mean = self.mean()
        var = self.var()
        m_2 = _mom(self.a, self.b, 2)
        m_3 = _mom(self.a, self.b, 3)
        m_4 = _mom(self.a, self.b, 4)
        return (m_4 + mean * (-4 * m_3 + mean * (6 * m_2 - 3 * mean**2))) / var**2 - 3

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return self.ppf(random_state.random(size))

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample, **kwargs):
        optimize_ml(self, sample, **kwargs)


@nb.njit(cache=True)
def nb_cdf(x, a, b):
    prob = 1 - (1 - x**a) ** b
    return cdf_bounds(prob, x, 0, 1)


@nb.njit(cache=True)
def nb_ppf(q, a, b):
    x_val = (1 - (1 - q) ** (1 / b)) ** (1 / a)
    return ppf_bounds_cont(x_val, q, 0, 1)


@nb.njit(cache=True)
def nb_entropy(a, b):
    h_b = digamma(b + 1) + np.euler_gamma
    return (1 - 1 / b) + (1 - 1 / a) * h_b - np.log(a) - np.log(b)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, a, b):
    if x <= 0 or x >= 1:
        return -np.inf
    else:
        return np.log(a * b) + xlogy((a - 1), x) + xlog1py((b - 1), -(x**a))


@nb.njit(cache=True)
def nb_neg_logpdf(x, a, b):
    return -(nb_logpdf(x, a, b)).sum()


def _mom(a, b, n):
    return b * beta(1 + n / a, b)



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from ..internal.optimization import optimize_ml
from ..internal.distribution_helper import all_not_none, eps
from .distributions import Continuous


class Logistic(Continuous):
    r"""
    Logistic distribution.

    The pdf of this distribution is

    .. math::

        f(x \mid \mu, s) =
            \frac{ \exp ( - \frac{x - \mu}{s})}{s(1 + \exp ( - \frac{x - \mu}{s}))^2}

    .. plot::
        :context: close-figs


        from preliz import Logistic, style
        style.use('preliz-doc')
        mus = [0., 0., -2.]
        ss = [1., 2., .4]
        for mu, s in zip(mus, ss):
            Logistic(mu, s).plot_pdf(support=(-5,5))

    =========  ==========================================
    Support    :math:`x \in \mathbb{R}`
    Mean       :math:`\mu`
    Variance   :math:`\frac{s^2 \pi^2}{3}`
    =========  ==========================================

    Parameters
    ----------
    mu : float
        Mean.
    s : float
        Scale (s > 0).
    """

    def __init__(self, mu=None, s=None):
        super().__init__()
        self.support = (-np.inf, np.inf)
        self._parametrization(mu, s)

    def _parametrization(self, mu=None, s=None):
        self.mu = mu
        self.s = s
        self.params = (self.mu, self.s)
        self.param_names = ("mu", "s")
        self.params_support = ((-np.inf, np.inf), (eps, np.inf))
        if all_not_none(self.mu, self.s):
            self._update(self.mu, self.s)

    def _update(self, mu, s):
        self.mu = np.float64(mu)
        self.s = np.float64(s)
        self.params = (self.mu, self.s)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.mu, self.s))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.mu, self.s)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.mu, self.s)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.s)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.s)

    def entropy(self):
        return nb_entropy(self.s)

    def mean(self):
        return self.mu

    def mode(self):
        return self.mu

    def median(self):
        return self.mu

    def var(self):
        return self.s**2 * np.pi**2 / 3

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 0

    def kurtosis(self):
        return 6 / 5

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.logistic(self.mu, self.s, size)

    def _fit_moments(self, mean, sigma):
        s = (3 * sigma**2 / np.pi**2) ** 0.5
        self._update(mean, s)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.njit(cache=True)
def nb_cdf(x, mu, s):
    return 1 / (1 + np.exp(-(x - mu) / s))


@nb.njit(cache=True)
def nb_ppf(q, mu, s):
    return mu + s * np.log(q / (1 - q))


@nb.njit(cache=True)
def nb_entropy(s):
    return np.log(s) + 2


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, s):
    if x == -np.inf:
        return -np.inf
    else:
        return -np.log(s) - 2 * np.log1p(np.exp(-(x - mu) / s)) - (x - mu) / s


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, s):
    return -(nb_logpdf(x, mu, s)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Discrete
from ..internal.distribution_helper import eps, all_not_none, num_skewness, num_kurtosis
from ..internal.optimization import optimize_moments, optimize_ml
from ..internal.special import cdf_bounds, ppf_bounds_disc


class DiscreteWeibull(Discrete):
    R"""
    Discrete Weibull distribution.

    The pmf of this distribution is

    .. math::

        f(x \mid q, \beta) = q^{x^{\beta}} - q^{(x+1)^{\beta}}

    .. plot::
        :context: close-figs


        from preliz import DiscreteWeibull, style
        style.use('preliz-doc')
        qs = [0.1, 0.9, 0.9]
        betas = [0.5, 0.5, 2]
        for q, b in zip(qs, betas):
            DiscreteWeibull(q, b).plot_pdf(support=(0,10))

    ========  ===============================================
    Support   :math:`x \in \mathbb{N}_0`
    Mean      :math:`\mu = \sum_{x = 1}^{\infty} q^{x^{\beta}}`
    Variance  :math:`2 \sum_{x = 1}^{\infty} x q^{x^{\beta}} - \mu - \mu^2`
    ========  ===============================================

    Parameters
    ----------
    q: float
        Shape parameter (0 < q < 1).
    beta: float
        Shape parameter (beta > 0).
    """

    def __init__(self, q=None, beta=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(q, beta)

    def _parametrization(self, q=None, beta=None):
        self.q = q
        self.beta = beta
        self.params = (self.q, self.beta)
        self.param_names = ("q", "beta")
        self.params_support = ((eps, 1 - eps), (eps, np.inf))
        if all_not_none(q, beta):
            self._update(q, beta)

    def _update(self, q, beta):
        self.q = np.float64(q)
        self.beta = np.float64(beta)
        self.params = (self.q, self.beta)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.q, self.beta))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.q, self.beta, self.support[0], self.support[1])

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.q, self.beta, self.support[0], self.support[1])

    def logpdf(self, x):
        return nb_logpdf(x, self.q, self.beta)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.q, self.beta)

    def entropy(self):
        x = self.xvals("full", 5000)
        logpdf = self.logpdf(x)
        return -np.sum(np.exp(logpdf) * logpdf)

    def mean(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.sum(x_values * pdf)

    def median(self):
        return self.ppf(0.5)

    def var(self):
        x_values = self.xvals("full")
        pdf = self.pdf(x_values)
        return np.sum((x_values - self.mean()) ** 2 * pdf)

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return num_skewness(self)

    def kurtosis(self):
        return num_kurtosis(self)

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return self.ppf(random_state.uniform(size=size))

    def _fit_moments(self, mean, sigma):
        optimize_moments(self, mean, sigma)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.vectorize(nopython=True, cache=True)
def nb_pdf(x, q, beta):
    if x < 0:
        return 0
    else:
        return q ** (x**beta) - q ** ((x + 1) ** beta)


@nb.njit(cache=True)
def nb_cdf(x, q, beta, lower, upper):
    prob = 1 - q ** ((x + 1) ** beta)
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(p, q, beta, lower, upper):
    x_val = np.ceil((np.log(1 - p) / np.log(q)) ** (1 / beta) - 1)
    return ppf_bounds_disc(x_val, p, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, q, beta):
    if x < 0:
        return -np.inf
    else:
        return np.log(q ** (x**beta) - q ** ((x + 1) ** beta))


@nb.njit(cache=True)
def nb_neg_logpdf(x, q, beta):
    return -(nb_logpdf(x, q, beta)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numpy as np
import numba as nb

from ..internal.optimization import optimize_ml
from ..internal.special import ppf_bounds_cont, xlogy
from ..internal.distribution_helper import all_not_none, eps
from .distributions import Continuous


class Pareto(Continuous):
    r"""
    Pareto distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \alpha, m) = \frac{\alpha m^{\alpha}}{x^{\alpha+1}}

    .. plot::
        :context: close-figs


        from preliz import Pareto, style
        style.use('preliz-doc')
        alphas = [1., 5., 5.]
        ms = [1., 1., 2.]
        for alpha, m in zip(alphas, ms):
            Pareto(alpha, m).plot_pdf(support=(0,4))

    ========  =============================================================
    Support   :math:`x \in [m, \infty)`
    Mean      :math:`\dfrac{\alpha m}{\alpha - 1}` for :math:`\alpha \ge 1`
    Variance  :math:`\dfrac{m^2 \alpha}{(\alpha - 1)^2 (\alpha - 2)}` for :math:`\alpha > 2`
    ========  =============================================================

    Parameters
    ----------
    alpha : float
        Shape parameter (alpha > 0).
    m : float
        Scale parameter (m > 0).
    """

    def __init__(self, alpha=None, m=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(alpha, m)

    def _parametrization(self, alpha=None, m=None):
        self.alpha = alpha
        self.m = m
        self.params = (self.alpha, self.m)
        self.param_names = ("alpha", "m")
        self.params_support = ((eps, np.inf), (eps, np.inf))
        if all_not_none(alpha, m):
            self._update(alpha, m)

    def _update(self, alpha, m):
        self.alpha = np.float64(alpha)
        self.m = np.float64(m)
        self.support = (self.m, np.inf)
        self.params = (self.alpha, self.m)
        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return np.exp(nb_logpdf(x, self.alpha, self.m))

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.alpha, self.m)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.alpha, self.m, 1, np.inf)

    def logpdf(self, x):
        return nb_logpdf(x, self.alpha, self.m)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.alpha, self.m)

    def entropy(self):
        return nb_entropy(self.alpha, self.m)

    def mean(self):
        return np.where(self.alpha > 1, self.alpha * self.m / (self.alpha - 1), np.inf)

    def median(self):
        return self.m * 2 ** (1 / self.alpha)

    def var(self):
        return np.where(
            self.alpha > 2,
            self.m**2 * self.alpha / ((self.alpha - 1) ** 2 * (self.alpha - 2)),
            np.inf,
        )

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return np.where(
            self.alpha > 3,
            2 * (1 + self.alpha) / (self.alpha - 3) * (1 - 2 / self.alpha) ** 0.5,
            np.nan,
        )

    def kurtosis(self):
        return np.where(
            self.alpha > 4,
            6
            * (self.alpha**3 + self.alpha**2 - 6 * self.alpha - 2)
            / (self.alpha * (self.alpha - 3) * (self.alpha - 4)),
            np.nan,
        )

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        random_samples = random_state.uniform(0, 1, size)
        return nb_rvs(random_samples, self.alpha, self.m)

    def _fit_moments(self, mean, sigma):
        alpha = 1 + (1 + (mean / sigma) ** 2) ** (1 / 2)
        m = (alpha - 1) * mean / alpha
        self._update(alpha, m)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


@nb.vectorize(nopython=True, cache=True)
def nb_cdf(x, alpha, m):
    if x < m:
        return 0
    return 1 - (m / x) ** alpha


@nb.njit(cache=True)
def nb_ppf(q, alpha, m, lower, upper):
    return ppf_bounds_cont(m * (1 - q) ** (-1 / alpha), q, lower, upper)


@nb.njit(cache=True)
def nb_entropy(alpha, m):
    return np.log((m / alpha) * np.exp(1 + 1 / alpha))


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, alpha, m):
    if x < m:
        return -np.inf
    return np.log(alpha) + xlogy(alpha, m) - xlogy((alpha + 1), x)


@nb.njit(cache=True)
def nb_neg_logpdf(x, alpha, m):
    return -(nb_logpdf(x, alpha, m)).sum()


@nb.njit(cache=True)
def nb_rvs(random_samples, alpha, m):
    return m / (1 - random_samples) ** (1 / alpha)



# pylint: disable=redefined-outer-name

import pytest
from pymc import Model
from numpy.testing import assert_almost_equal
import numpy as np

from preliz.distributions import (
    AsymmetricLaplace,
    Beta,
    BetaScaled,
    Cauchy,
    ChiSquared,
    Gamma,
    Gumbel,
    ExGaussian,
    Exponential,
    HalfCauchy,
    HalfNormal,
    HalfStudentT,
    InverseGamma,
    Kumaraswamy,
    Laplace,
    Logistic,
    LogLogistic,
    LogNormal,
    LogitNormal,
    Moyal,
    Normal,
    Pareto,
    Rice,
    SkewNormal,
    StudentT,
    Triangular,
    TruncatedNormal,
    Uniform,
    VonMises,
    Wald,
    Weibull,
    Bernoulli,
    BetaBinomial,
    Binomial,
    DiscreteUniform,
    DiscreteWeibull,
    Geometric,
    NegativeBinomial,
    Poisson,
    ZeroInflatedNegativeBinomial,
    ZeroInflatedPoisson,
    Dirichlet,
    MvNormal,
    Truncated,
    Censored,
    Hurdle,
)


@pytest.fixture(scope="session")
def a_few_poissons():
    return Poisson(), Poisson([1.0, 2.0]), Poisson(4.5)


@pytest.fixture(scope="session")
def multivariates():
    return Dirichlet(), Dirichlet([1.0, 2.0, 1.0]), MvNormal(np.zeros(2), np.eye(2))


@pytest.mark.parametrize(
    "distribution, params",
    [
        (AsymmetricLaplace, (1, 0, 1)),
        (Beta, (2, 5)),
        (ChiSquared, (1,)),
        (ExGaussian, (0, 1, 1e-6)),
        (Exponential, (0.5,)),
        (Gamma, (1, 0.5)),
        (Gumbel, (1, 2)),
        (HalfNormal, (1,)),
        (HalfStudentT, (100, 1)),
        (InverseGamma, (3, 1)),
        (Kumaraswamy, (5, 2)),
        (Laplace, (0, 1)),
        (Logistic, (1, 2)),
        (LogLogistic, (1, 5)),
        (LogNormal, (0, 0.5)),
        (LogitNormal, (0, 0.5)),
        (Moyal, (1, 2)),
        (Normal, (0, 1)),
        (Pareto, (5, 1)),
        (Rice, (4, 1)),
        (SkewNormal, (0, 1, 0)),
        (StudentT, (100, 0, 1)),
        (Triangular, (-2, 3, 7)),
        (TruncatedNormal, (0, 1, -np.inf, np.inf)),
        (Uniform, (0, 1)),
        (VonMises, (0, 1000)),
        (Wald, (1, 1)),
        (Weibull, (2, 1)),
        (Bernoulli, (0.8,)),
        (BetaBinomial, (1, 1, 10)),
        (Binomial, (2, 0.5)),
        (Binomial, (2, 0.1)),
        (DiscreteUniform, (0, 1)),
        (DiscreteWeibull, (0.5, 3)),
        (Geometric, (0.75,)),
        (NegativeBinomial, (8, 4)),
        (Poisson, (4.5,)),
        (
            ZeroInflatedPoisson,
            (
                0.8,
                4.5,
            ),
        ),
    ],
)
def test_moments(distribution, params):
    dist = distribution(*params)
    dist_ = distribution()

    dist_._fit_moments(dist.mean(), dist.std())

    tol = 5
    if dist.__class__.__name__ in [
        "BetaBinomial",
        "Binomial",
        "DiscreteWeibull",
        "ExGaussian",
        "NegativeBinomial",
        "Kumaraswamy",
        "LogitNormal",
        "Rice",
        "StudentT",
    ]:
        tol = 0
    assert_almost_equal(dist.mean(), dist_.mean(), tol)
    assert_almost_equal(dist.std(), dist_.std(), tol)
    assert_almost_equal(params, dist_.params, 0)


@pytest.mark.parametrize(
    "distribution, params",
    [
        (AsymmetricLaplace, (1, 4, 3)),
        (Beta, (2, 5)),
        (BetaScaled, (2, 2, -1, 2)),
        (Cauchy, (0, 1)),
        (ChiSquared, (1,)),
        (ExGaussian, (0, 1, 3)),
        (Exponential, (0.5,)),
        (Gamma, (1, 0.5)),
        (Gumbel, (0, 1)),
        (HalfCauchy, (1,)),
        (HalfNormal, (1,)),
        (HalfStudentT, (100, 1)),
        (InverseGamma, (3, 0.5)),
        (Kumaraswamy, (2, 3)),
        (Laplace, (0, 1)),
        (Logistic, (0, 1)),
        (LogNormal, (0, 0.5)),
        (LogitNormal, (0, 0.5)),
        (Moyal, (0, 2)),
        (Normal, (0, 1)),
        (Pareto, (5, 1)),
        (Rice, (1, 2)),
        (SkewNormal, (0, 1, 0)),
        (SkewNormal, (0, 1, -1)),
        (StudentT, (4, 0, 1)),
        (StudentT, (1000, 0, 1)),
        (Triangular, (-3, 0, 5)),
        (TruncatedNormal, (0, 1, -1, 1)),
        (Uniform, (0, 1)),
        (VonMises, (0, 1)),
        (Wald, (1, 1)),
        (Weibull, (2, 1)),
        (Bernoulli, (0.4,)),
        (BetaBinomial, (2, 2, 10)),
        (Binomial, (2, 0.5)),
        (Binomial, (2, 0.1)),
        (DiscreteUniform, (0, 1)),
        (DiscreteWeibull, (0.1, 1.5)),
        (Geometric, (0.5,)),
        (NegativeBinomial, (8, 4)),
        (Poisson, (4.5,)),
        (ZeroInflatedNegativeBinomial, (0.7, 8, 4)),
        (
            ZeroInflatedPoisson,
            (
                0.8,
                4.5,
            ),
        ),
    ],
)
def test_mle(distribution, params):
    for _ in range(10):
        dist = distribution(*params)
        sample = dist.rvs(20_000)
        dist_ = distribution()
        dist_._fit_mle(sample)

        if dist.__class__.__name__ in ["Pareto", "ExGaussian"]:
            tol = 0
        else:
            tol = 1
        try:
            assert_almost_equal(dist.mean(), dist_.mean(), tol)
            assert_almost_equal(dist.std(), dist_.std(), tol)
            if dist.__class__.__name__ == "StudentT":
                assert_almost_equal(params[1:], dist_.params[1:], 0)
            else:
                assert_almost_equal(params, dist_.params, 0)
            break
        except AssertionError:
            pass
    else:
        raise AssertionError("Test failed after 10 attempts")


@pytest.mark.parametrize("fmt", (".2f", ".1g"))
@pytest.mark.parametrize("interval", ("hdi", "eti", [0.25, 0.75]))
@pytest.mark.parametrize("mass", (0.5, 0.95))
def test_summary_args(mass, interval, fmt):
    result = Normal(0, 1).summary(mass=mass, interval=interval, fmt=fmt)
    assert result.mean == 0
    assert result.std == 1


def test_summary_univariate_valid(a_few_poissons):
    d_0, d_1, d_2 = a_few_poissons
    with pytest.raises(ValueError):
        d_0.summary()
    with pytest.raises(ValueError):
        d_1.summary()
    result = d_2.summary()
    assert result.__class__.__name__ == "Poisson"
    assert result.mean == 4.5
    assert result.std == 2.12
    assert result.lower == 1.0
    assert result.upper == 8.0


def test_summary_multivariate_valid(multivariates):
    d_0, d_1, m_0 = multivariates
    with pytest.raises(ValueError):
        d_0.summary()
    result = d_1.summary()
    assert result.__class__.__name__ == "Dirichlet"
    assert_almost_equal(result.mean, [0.25, 0.5, 0.25])
    assert_almost_equal(result.std, [0.193, 0.223, 0.193], decimal=3)
    result = m_0.summary()
    assert result.__class__.__name__ == "MvNormal"
    assert_almost_equal(result.mean, [0, 0])
    assert_almost_equal(result.std, [1, 1])


def test_eti(a_few_poissons):
    d_0, d_1, d_2 = a_few_poissons
    with pytest.raises(ValueError):
        d_0.eti()
    with pytest.raises(ValueError):
        d_1.eti()
    result = d_2.eti()
    assert result == (1.0, 9.0)


def test_hdi(a_few_poissons):
    d_0, d_1, d_2 = a_few_poissons
    with pytest.raises(ValueError):
        d_0.hdi()
    with pytest.raises(ValueError):
        d_1.hdi()
    result = d_2.hdi()
    assert result == (1.0, 8.0)


def test_cdf(a_few_poissons):
    _, d_1, d_2 = a_few_poissons
    result1 = d_1.cdf(1)
    result2 = d_2.cdf(1)
    assert round(result2, 2) == 0.06
    assert np.allclose(result1, (0.73, 0.41), 2)


def test_ppf(a_few_poissons):
    _, d_1, d_2 = a_few_poissons
    result1 = d_1.ppf(0.5)
    result2 = d_2.ppf(0.5)
    assert np.allclose(result1, (1, 2))
    assert result2 == 4.0


def test_to_pymc():
    with Model() as model:
        Gamma(1, 1).to_pymc("a", shape=(2, 2))
        Hurdle(Gamma(1, 1), psi=0.5).to_pymc("b", shape=1)
        Truncated(Gamma(1, 1), lower=1).to_pymc("c")

    assert model.basic_RVs[0].name == "a"
    assert model.basic_RVs[0].ndim == 2
    assert model.basic_RVs[1].name == "b"
    assert model.basic_RVs[1].ndim == 1
    assert model.basic_RVs[2].name == "c"
    assert model.basic_RVs[2].ndim == 0
    assert Normal(0, 1).to_pymc(shape=2).ndim == 1
    assert Censored(Normal(0, 1), lower=0).to_pymc().ndim == 0


def test_to_bambi():
    bambi_prior = Gamma(mu=2, sigma=1).to_bambi()
    assert bambi_prior.name == "Gamma"
    assert bambi_prior.args["mu"] == 2
    assert bambi_prior.args["sigma"] == 1



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
# pylint: disable=invalid-name
import numba as nb
import numpy as np
from scipy.special import ndtr, expi  # pylint: disable=no-name-in-module

from .distributions import Continuous
from ..internal.distribution_helper import eps, all_not_none
from ..internal.special import cdf_bounds
from ..internal.optimization import optimize_ml, find_ppf


class Wald(Continuous):
    r"""
    Wald distribution.

    The pdf of this distribution is

    .. math::

       f(x \mid \mu, \lambda) =
           \left(\frac{\lambda}{2\pi}\right)^{1/2} x^{-3/2}
           \exp\left\{
               -\frac{\lambda}{2x}\left(\frac{x-\mu}{\mu}\right)^2
           \right\}

    .. plot::
        :context: close-figs


        from preliz import Wald, style
        style.use('preliz-doc')
        mus = [1., 1.]
        lams = [1., 3.]
        for mu, lam in zip(mus, lams):
            Wald(mu, lam).plot_pdf(support=(0,4))

    ========  =============================
    Support   :math:`x \in (0, \infty)`
    Mean      :math:`\mu`
    Variance  :math:`\dfrac{\mu^3}{\lambda}`
    ========  =============================

    Wald distribution has 3 alternative parametrizations. In terms of mu and lam,
    mu and phi or lam and phi.

    The link between the 3 alternatives is given by

    .. math::

       \phi = \dfrac{\lambda}{\mu}

    Parameters
    ----------
    mu : float
        Mean of the distribution (mu > 0).
    lam : float
        Relative precision (lam > 0).
    phi : float
        Shape parameter (phi > 0).
    """

    def __init__(self, mu=None, lam=None, phi=None):
        super().__init__()
        self.support = (0, np.inf)
        self._parametrization(mu, lam, phi)

    def _parametrization(self, mu=None, lam=None, phi=None):
        if all_not_none(mu, lam, phi):
            raise ValueError(
                "Incompatible parametrization. Either use mu and lam or mu and phi or lam and phi."
            )

        self.param_names = ("mu", "lam")
        self.params_support = ((eps, np.inf), (eps, np.inf))

        if phi is not None:
            self.phi = phi
            if (mu and phi) is not None:
                lam = self._from_mu_phi(mu, phi)
                self.param_names = ("mu", "phi")

            elif (lam and phi) is not None:
                mu = self._from_lam_phi(lam, phi)
                self.param_names = ("lam", "phi")

        self.mu = mu
        self.lam = lam
        if all_not_none(self.mu, self.lam):
            self._update(self.mu, self.lam)

    def _from_mu_phi(self, mu, phi):
        lam = mu * phi
        return lam

    def _from_lam_phi(self, lam, phi):
        mu = lam / phi
        return mu

    def _to_phi(self, mu, lam):
        phi = lam / mu
        return phi

    def _update(self, mu, lam):
        self.mu = np.float64(mu)
        self.lam = np.float64(lam)
        self.phi = self._to_phi(self.mu, self.lam)

        if self.param_names == ("mu", "lam"):
            self.params = (self.mu, self.lam)
        elif self.param_names == ("mu", "phi"):
            self.params = (self.mu, self.phi)
        elif self.param_names == ("lam", "phi"):
            self.params = (self.lam, self.phi)

        self.is_frozen = True

    def pdf(self, x):
        x = np.array(x)
        return np.exp(self.logpdf(x))

    def cdf(self, x):
        x = np.array(x)
        return nb_cdf(x, self.mu, self.lam)

    def ppf(self, q):
        q = np.asarray(q)
        return find_ppf(self, q)

    def logpdf(self, x):
        return nb_logpdf(x, self.mu, self.lam)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.mu, self.lam)

    def entropy(self):
        return nb_entropy(self.mu, self.lam)

    def mean(self):
        return self.mu

    def median(self):
        return self.ppf(0.5)

    def var(self):
        return self.mu**3 / self.lam

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 3 * (self.mu / self.lam) ** 0.5

    def kurtosis(self):
        return 15 * self.mu / self.lam

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.wald(self.mu, self.lam, size)

    def _fit_moments(self, mean, sigma):
        lam = mean**3 / sigma**2
        self._update(mean, lam)

    def _fit_mle(self, sample):
        optimize_ml(self, sample)


def nb_cdf(x, mu, lam):
    x = np.asarray(x)
    u = (lam / (x + eps)) ** 0.5
    v = x / mu
    z = ndtr(u * (v - 1)) + np.exp(2 * lam / mu) * ndtr(-u * (v + 1))
    return cdf_bounds(z, x, 0, np.inf)


def nb_entropy(mu, lam):
    return 0.5 * np.log((2 * np.pi * np.e * mu**3) / lam) + 3 / 2 * np.exp(2 * lam / mu) * expi(
        -2 * lam / mu
    )


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, mu, lam):
    if x <= 0:
        return -np.inf
    elif x == np.inf:
        return -np.inf
    else:
        return (
            np.log(lam) - (np.log(2 * np.pi) + 3 * np.log(x)) - lam * (x - mu) ** 2 / (mu**2 * x)
        ) / 2


@nb.njit(cache=True)
def nb_neg_logpdf(x, mu, lam):
    return -(nb_logpdf(x, mu, lam)).sum()



# pylint: disable=attribute-defined-outside-init
# pylint: disable=arguments-differ
import numba as nb
import numpy as np

from .distributions import Continuous
from ..internal.distribution_helper import all_not_none
from ..internal.special import cdf_bounds, ppf_bounds_cont


class Uniform(Continuous):
    r"""
    Uniform distribution.

    The pdf of this distribution is

    .. math:: f(x \mid lower, upper) = \frac{1}{upper-lower}

    .. plot::
        :context: close-figs


        from preliz import Uniform, style
        style.use('preliz-doc')
        ls = [1, -2]
        us = [6, 2]
        for l, u in zip(ls, us):
            ax = Uniform(l, u).plot_pdf()
        ax.set_ylim(0, 0.3)

    ========  =====================================
    Support   :math:`x \in [lower, upper]`
    Mean      :math:`\dfrac{lower + upper}{2}`
    Variance  :math:`\dfrac{(upper - lower)^2}{12}`
    ========  =====================================

    Parameters
    ----------
    lower: float
        Lower limit.
    upper: float
        Upper limit (upper > lower).
    """

    def __init__(self, lower=None, upper=None):
        super().__init__()
        self._parametrization(lower, upper)

    def _parametrization(self, lower=None, upper=None):
        self.lower = lower
        self.upper = upper
        self.params = (self.lower, self.upper)
        self.param_names = ("lower", "upper")
        self.params_support = ((-np.inf, np.inf), (-np.inf, np.inf))
        if lower is None:
            self.lower = -np.inf
        if upper is None:
            self.upper = np.inf
        self.support = (self.lower, self.upper)
        if all_not_none(lower, upper):
            self._update(lower, upper)
        else:
            self.lower = lower
            self.upper = upper

    def _update(self, lower, upper):
        self.lower = np.float64(lower)
        self.upper = np.float64(upper)
        self.params = (self.lower, self.upper)
        self.support = (self.lower, self.upper)

        self.is_frozen = True

    def pdf(self, x):
        x = np.asarray(x)
        return nb_pdf(x, self.lower, self.upper)

    def cdf(self, x):
        x = np.asarray(x)
        return nb_cdf(x, self.lower, self.upper)

    def ppf(self, q):
        q = np.asarray(q)
        return nb_ppf(q, self.lower, self.upper)

    def logpdf(self, x):
        return nb_logpdf(x, self.lower, self.upper)

    def _neg_logpdf(self, x):
        return nb_neg_logpdf(x, self.lower, self.upper)

    def entropy(self):
        return nb_entropy(self.lower, self.upper)

    def mean(self):
        return (self.upper + self.lower) / 2

    def median(self):
        return (self.upper + self.lower) / 2

    def var(self):
        return (self.upper - self.lower) ** 2 / 12

    def std(self):
        return self.var() ** 0.5

    def skewness(self):
        return 0

    def kurtosis(self):
        return -6 / 5

    def rvs(self, size=None, random_state=None):
        random_state = np.random.default_rng(random_state)
        return random_state.uniform(self.lower, self.upper, size)

    def _fit_moments(self, mean, sigma):
        lower = mean - 1.73205 * sigma
        upper = mean + 1.73205 * sigma
        self._update(lower, upper)

    def _fit_mle(self, sample):
        lower = np.min(sample)
        upper = np.max(sample)
        self._update(lower, upper)


@nb.njit(cache=True)
def nb_cdf(x, lower, upper):
    prob = (x - lower) / (upper - lower)
    return cdf_bounds(prob, x, lower, upper)


@nb.njit(cache=True)
def nb_ppf(q, lower, upper):
    x_vals = lower + q * (upper - lower)
    return ppf_bounds_cont(x_vals, q, lower, upper)


@nb.vectorize(nopython=True, cache=True)
def nb_pdf(x, lower, upper):
    if lower <= x <= upper:
        return 1 / (upper - lower)
    else:
        return 0


@nb.njit(cache=True)
def nb_entropy(lower, upper):
    return np.log(upper - lower)


@nb.vectorize(nopython=True, cache=True)
def nb_logpdf(x, lower, upper):
    if lower <= x <= upper:
        return -np.log(upper - lower)
    else:
        return -np.inf


@nb.njit(cache=True)
def nb_neg_logpdf(x, lower, upper):
    return -(nb_logpdf(x, lower, upper)).sum()
